# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=R0904

from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from ..data import CommentsData, Data
from .entity import Entity


if TYPE_CHECKING:
    from .. import shared
    from ..groups import PropertyGroup

DEFAULT_CRS = {"Code": "Unknown", "Name": "Unknown"}


class EntityContainer(Entity):
    """
    Base Entity class
    """

    _TYPE_UID: uuid.UUID | None = None

    def __init__(self, **kwargs):
        self._children: list = []
        super().__init__(**kwargs)

    @classmethod
    def default_type_uid(cls) -> uuid.UUID | None:
        """
        Default uuid for the entity type.
        """
        return cls._TYPE_UID

    def add_children(
        self, children: Entity | PropertyGroup | list[Entity | PropertyGroup]
    ):
        """
        :param children: Add a list of entities as
            :obj:`~geoh5py.shared.entity.Entity.children`
        """
        if not isinstance(children, list):
            children = [children]

        for child in children:
            if child in self._children:
                continue

            if not isinstance(child, Entity):
                raise TypeError(
                    f"Child must be an instance of Entity, not {type(child)}"
                )

            self._children.append(child)

            if hasattr(child, "parent") and child.parent is not self:
                child.parent = self

    def add_comment(self, comment: str, author: str | None = None):
        """
        Add text comment to an object.

        :param comment: Text to be added as comment.
        :param author: Author's name or :obj:`~geoh5py.workspace.workspace.Workspace.contributors`.
        """

        date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        if author is None:
            author = ",".join(self.workspace.contributors)

        comment_dict = {"Author": author, "Date": date, "Text": comment}

        if self.comments is None:
            self.workspace.create_entity(
                Data,
                entity={
                    "name": "UserComments",
                    "association": "OBJECT",
                    "values": {"Comments": [comment_dict]},
                    "parent": self,
                },
                entity_type={"primitive_type": "TEXT"},
            )

        else:
            self.comments.values = {
                "Comments": self.comments.values["Comments"] + [comment_dict]
            }

    def add_file(self, file: str | Path | bytes, name: str = "filename.dat"):
        """
        Add a file to the object or group stored as bytes on a FilenameData

        :param file: File name with path to import.
        :param name: Name of the file in the workspace.
        """
        if isinstance(file, str):
            file = Path(file)

        if isinstance(file, Path):
            if not file.is_file():
                raise ValueError(f"Input file '{file}' does not exist.")

            name = Path(file).name

            with open(file, "rb") as raw_binary:
                blob = raw_binary.read()

        elif isinstance(file, bytes):
            blob = file

        else:
            raise TypeError(
                f"Input file must be a path or BytesIO object, not {type(file)}"
            )

        attributes = {
            "name": name,
            "file_bytes": blob,
            "association": "OBJECT",
            "parent": self,
            "values": name,
        }
        entity_type = {"name": "UserFiles", "primitive_type": "FILENAME"}

        file_data = self.workspace.create_entity(
            None, entity=attributes, entity_type=entity_type
        )

        return file_data

    @property
    def children(self):
        """
        :obj:`list` Children entities in the workspace tree
        """
        return self._children

    @property
    def comments(self):
        """
        Fetch a :obj:`~geoh5py.data.data.CommentsData` entity from children.
        """
        for child in self.children:
            if isinstance(child, CommentsData):
                return child

        return None

    @abstractmethod
    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        **kwargs,
    ):
        """
        Function to copy an entity to a different parent entity.

        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param copy_children: (Optional) Create copies of all children entities along with it.
        :param clear_cache: Clear array attributes after copy to minimize the
            memory footprint of the workspace.
        :param kwargs: Additional keyword arguments to pass to the copy constructor.

        :return entity: Registered Entity to the workspace.
        """

    def copy_from_extent(
        self,
        extent: np.ndarray,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> Entity | None:
        """
        Function to copy an entity to a different parent entity.

        :param extent: Bounding box extent requested for the input entity, as supplied for
            :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.
        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param copy_children: (Optional) Create copies of all children entities along with it.
        :param clear_cache: Clear array attributes after copy.
        :param inverse: Keep the inverse (clip) of the extent selection.
        :param kwargs: Additional keyword arguments to pass to the copy constructor.

        :return entity: Registered Entity to the workspace.
        """
        indices = self.mask_by_extent(extent, inverse=inverse)
        if indices is None:
            return None

        return self.copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=indices,
            **kwargs,
        )

    def get_entity(self, name: str | uuid.UUID) -> list[Entity | None]:
        """
        Get a child :obj:`~geoh5py.data.data.Data` by name.

        :param name: Name of the target child data
        :param entity_type: Sub-select entities based on type.
        :return: A list of children Data objects
        """

        if isinstance(name, uuid.UUID):
            entity_list = [child for child in self.children if child.uid == name]
        else:
            entity_list = [child for child in self.children if child.name == name]

        if not entity_list:
            return [None]

        return entity_list

    def get_entity_list(self, entity_type=ABC) -> list[str]:
        """
        Get a list of names of all children :obj:`~geoh5py.data.data.Data`.

        :param entity_type: Option to sub-select based on type.
        :return: List of names of data associated with the object.
        """
        name_list = [
            child.name for child in self.children if isinstance(child, entity_type)
        ]
        return sorted(name_list)

    def remove_children(self, children: list[shared.Entity | PropertyGroup]):
        """
        Remove children from the list of children entities.

        :param children: List of entities

        .. warning::
            Removing a child entity without re-assigning it to a different
            parent may cause it to become inactive. Inactive entities are removed
            from the workspace by
            :func:`~geoh5py.shared.weakref_utils.remove_none_referents`.
        """
        if not isinstance(children, list):
            children = [children]

        self._children = [child for child in self._children if child not in children]
        self.workspace.remove_children(self, children)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=no-member, too-many-lines, too-many-ancestors
# mypy: disable-error-code="attr-defined"

from __future__ import annotations

import json
import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import numpy as np

from geoh5py.data import FloatData, IntegerData, ReferencedData
from geoh5py.groups.property_group import PropertyGroup
from geoh5py.objects import Curve
from geoh5py.objects.surveys.base import BaseSurvey
from geoh5py.shared.utils import str2uuid, str_json_to_dict


if TYPE_CHECKING:
    from geoh5py.groups import Group
    from geoh5py.workspace import Workspace


class BaseEMSurvey(BaseSurvey, ABC):  # pylint: disable=too-many-public-methods
    """
    A base electromagnetics survey object.
    """

    __INPUT_TYPE = None
    __OMIT_LIST = (
        "_receivers",
        "_transmitters",
        "_base_stations",
        "_tx_id_property",
        "_metadata",
    )
    __TYPE_MAP = {
        "Transmitters": "transmitters",
        "Receivers": "receivers",
        "Base stations": "base_stations",
    }
    __UNITS = None

    def __init__(self, **kwargs):
        self._receivers: BaseEMSurvey | None = None
        self._transmitters: BaseEMSurvey | None = None
        self._tx_id_property: ReferencedData | IntegerData | None = None

        super().__init__(**kwargs)

    def add_components_data(self, data: dict) -> list[PropertyGroup]:
        """
        Add lists of data components to an EM survey. The name of each component is
        appended to the metadata 'Property groups'.

        Data channels must be provided for every frequency or time
        in order specified by
        :attr:`~geoh5py.objects.surveys.electromagnetics.BaseEMSurvey.channels`.
        The data channels can be supplied as either a list of
        :obj:`geoh5py.data.float_data.FloatData` entities or :obj:`uuid.UUID`

        .. code-block:: python

            data = {
                "Component A": [
                    data_entity_1,
                    data_entity_2,
                ],
                "Component B": [...],
            },


        or a nested dictionary of arguments defining new Data entities as defined by the
        :func:`~geoh5py.objects.object_base.ObjectBase.add_data` method.

        .. code-block:: python

            data = {
                "Component A": {
                    time_1: {
                        'values': [v_11, v_12, ...],
                        "entity_type": entity_type_A,
                        ...,
                    },
                    time_2: {...},
                    ...,
                },
                "Component B": {...},
            }

        :param data: Dictionary of data components to be added to the survey.

        :return: List of property groups for all components added.
        """
        prop_groups = []
        if self.channels is None or not self.channels:
            raise AttributeError(
                "The 'channels' attribute of an EMSurvey class must be set before the "
                "'add_components_data' method can be used."
            )

        if not isinstance(data, dict):
            raise TypeError(
                "Input data must be nested dictionaries of components and channels."
            )

        for name, data_block in data.items():
            prop_group = self.add_validate_component_data(name, data_block)
            prop_groups.append(prop_group)

        return prop_groups

    def add_validate_component_data(self, name: str, data_block: list | dict):
        """
        Append a property group to the entity and its metadata after validations.
        """
        if self.property_groups is not None and name in [
            pg.name for pg in self.property_groups
        ]:
            raise ValueError(
                f"PropertyGroup named '{name}' already exists on the survey entity. "
                f"Consider using the 'edit_em_metadata' method with "
                "'Property groups' argument instead."
            )

        if not isinstance(data_block, (dict, list)) or (
            isinstance(data_block, list)
            and not all(isinstance(entry, FloatData) for entry in data_block)
        ):
            raise TypeError(
                f"List of values provided for component '{name}' must be a list "
                f"of {FloatData} or {dict} of attributes. "
                f"Values of type {type(data_block)} provided."
            )

        if len(data_block) != len(self.channels):
            raise ValueError(
                f"The number of channel values provided must be of len({len(self.channels)}) "
                "corresponding to the 'channels' attribute. "
                f"Value of {type(data_block)} and len({len(data_block)}) provided."
            )

        if isinstance(data_block, list):
            if not np.all([entry.parent == self for entry in data_block]):
                raise ValueError(
                    f"The list of values provided for the component '{name}' "
                    f"must contain {FloatData} belonging to the target survey."
                )

            data_list = data_block

        else:
            data_list = []
            for channel, attr in data_block.items():
                if not isinstance(attr, dict):
                    raise TypeError(
                        f"Given value to data {channel} should of type {dict} or attributes. "
                        f"Type {type(attr)} given instead."
                    )
                data_list.append(self.add_data({channel: attr}))

        prop_group = self.add_data_to_group(data_list, name)
        self.edit_em_metadata({"Property groups": prop_group})

        return prop_group

    @property
    def channels(self):
        """
        List of measured channels.
        """
        channels = self.metadata["EM Dataset"]["Channels"]
        return channels

    @channels.setter
    def channels(self, values: list | np.ndarray):
        if isinstance(values, np.ndarray):
            values = values.tolist()

        if not isinstance(values, list) or not np.all(
            [isinstance(x, float) for x in values]
        ):
            raise TypeError(
                f"Values provided as 'channels' must be a list of {float}. {type(values)} provided"
            )

        self.edit_em_metadata({"Channels": values})

    @property
    def complement(self) -> BaseEMSurvey | None:
        """Returns the complement object for self."""
        return None

    @property
    def complement_reference(self):
        """Reference data linking the geometry of complement entity."""
        return self.tx_id_property

    @complement_reference.setter
    def complement_reference(
        self, value: ReferencedData | IntegerData | np.ndarray | None
    ):
        self.tx_id_property = value

    @property
    def components(self) -> dict | None:
        """
        Rapid access to the list of data entities for all components.
        """
        if "Property groups" in self.metadata["EM Dataset"]:
            components = {}
            for name in self.metadata["EM Dataset"]["Property groups"]:
                prop_group = self.fetch_property_group(name=name)

                if prop_group.properties is None:
                    continue

                components[name] = [
                    self.workspace.get_entity(uid)[0] for uid in prop_group.properties
                ]
            return components

        return None

    @property
    @abstractmethod
    def default_input_types(self) -> list[str] | None:
        """
        Input types.

        Must be one of 'Rx', 'Tx', 'Tx and Rx', 'Rx only', 'Rx and base stations'."""

    @property
    def default_metadata(self) -> dict:
        """Default metadata structure. Implemented on the child class."""
        return {"EM Dataset": {}}

    @property
    @abstractmethod
    def default_transmitter_type(self) -> type:
        """
        :return: Transmitters implemented on the child class.
        """

    @property
    @abstractmethod
    def default_receiver_type(self) -> type:
        """
        :return: Receivers implemented on the child class.
        """

    @property
    @abstractmethod
    def default_units(self) -> list[str]:
        """
        List of accepted units.
        """

    def edit_em_metadata(self, entries: dict[str, Any]):
        """
        Utility function to edit or add metadata fields and trigger an update
        on the receiver and transmitter entities.

        :param entries: Metadata key value pairs.
        """
        em_metadata = self.metadata.get("EM Dataset", {})

        for key, value in entries.items():
            if key == "Property groups":
                self._edit_validate_property_groups(value)
            elif value is None:
                if key in em_metadata:
                    del em_metadata[key]
            else:
                em_metadata[key] = value

        self.metadata = {"EM Dataset": em_metadata}

    @property
    def input_type(self) -> str | None:
        """Data input type. Must be one of 'Rx', 'Tx' or 'Tx and Rx'"""
        if "Input type" in self.metadata["EM Dataset"]:
            return self.metadata["EM Dataset"]["Input type"]

        return None

    @input_type.setter
    def input_type(self, value: str):
        if self.default_input_types is None:
            return

        if value not in self.default_input_types:
            raise ValueError(
                "Input 'input_type' must be one of "
                f"{self.default_input_types}. {value} provided."
            )
        self.edit_em_metadata({"Input type": value})

    @property
    def metadata(self):
        """Metadata attached to the entity."""
        if getattr(self, "_metadata", None) is None:
            metadata = self.workspace.fetch_metadata(self)
            self._metadata = self.validate_em_metadata(metadata)

        return self._metadata

    @metadata.setter
    def metadata(self, values: dict | np.ndarray | bytes | None):
        self._metadata = self.validate_em_metadata(values)

        if self.on_file:
            self.workspace.update_attribute(self, "metadata")

            if self.complement is not None:
                self.complement._metadata = self._metadata  # pylint: disable=protected-access

                if self.complement.on_file:
                    self.workspace.update_attribute(self.complement, "metadata")

    @property
    def receivers(self) -> BaseEMSurvey | None:
        """
        The associated TEM receivers.
        """
        if getattr(self, "_receivers", None) is None:
            if self.metadata is not None and "Receivers" in self.metadata["EM Dataset"]:
                receiver = self.metadata["EM Dataset"]["Receivers"]
                receiver_entity = self.workspace.get_entity(receiver)[0]

                if isinstance(receiver_entity, BaseEMSurvey):
                    self._receivers = receiver_entity

        return self._receivers

    @receivers.setter
    def receivers(self, receivers: BaseEMSurvey):
        if not isinstance(receivers, self.default_receiver_type):
            raise TypeError(
                f"Provided receivers must be of type {self.default_receiver_type}. "
                f"{type(receivers)} provided."
            )

        if receivers.tx_id_property is not None:
            self.edit_em_metadata({"Tx ID property": receivers.tx_id_property.uid})

            if isinstance(
                self.tx_id_property, ReferencedData | IntegerData
            ) and isinstance(receivers.tx_id_property, ReferencedData | IntegerData):
                self.tx_id_property.entity_type = receivers.tx_id_property.entity_type

        self._receivers = receivers
        self.edit_em_metadata({"Receivers": receivers.uid})

    @property
    def survey_type(self) -> str | None:
        """Data input type. Must be one of 'Rx', 'Tx' or 'Tx and Rx'"""
        if "Survey type" in self.metadata["EM Dataset"]:
            return self.metadata["EM Dataset"]["Survey type"]

        return None

    @property
    def transmitters(self):
        """
        The associated TEM transmitters (sources).
        """
        if getattr(self, "_transmitters", None) is None:
            if (
                self.metadata is not None
                and "Transmitters" in self.metadata["EM Dataset"]
            ):
                transmitter = self.metadata["EM Dataset"]["Transmitters"]
                transmitter_entity = self.workspace.get_entity(transmitter)[0]

                if isinstance(transmitter_entity, BaseEMSurvey):
                    self._transmitters = transmitter_entity

        return self._transmitters

    @transmitters.setter
    def transmitters(self, transmitters: BaseEMSurvey):
        if isinstance(None, self.default_transmitter_type):
            raise AttributeError(
                f"The 'transmitters' attribute cannot be set on class {type(self)}."
            )

        if not isinstance(transmitters, self.default_transmitter_type):
            raise TypeError(
                f"Provided transmitters must be of type {self.default_transmitter_type}. "
                f"{type(transmitters)} provided."
            )

        if transmitters.tx_id_property is not None:
            self.edit_em_metadata(
                {"Tx ID tx property": transmitters.tx_id_property.uid}
            )
            if isinstance(
                self.tx_id_property, ReferencedData | IntegerData
            ) and isinstance(transmitters.tx_id_property, ReferencedData | IntegerData):
                self.tx_id_property.entity_type = (
                    transmitters.tx_id_property.entity_type
                )

        self._transmitters = transmitters
        self.edit_em_metadata({"Transmitters": transmitters.uid})

    @property
    def omit_list(self) -> tuple:
        """
        List of attributes to omit when copying.
        """
        return self.__OMIT_LIST

    @property
    def type_map(self) -> dict[str, str]:
        """
        Mapping of the electrode types to the associated electrode.
        """
        return self.__TYPE_MAP

    @property
    def tx_id_property(self) -> ReferencedData | IntegerData | None:
        """
        Data link between the receiver and transmitter object.
        """
        if self._tx_id_property is None and self.metadata is not None:
            self._tx_id_property = self._fetch_transmitter_id()

        return self._tx_id_property

    @tx_id_property.setter
    def tx_id_property(self, value: uuid.UUID | ReferencedData | np.ndarray | None):
        if isinstance(value, uuid.UUID):
            value = self.get_data(value)[0]

        if isinstance(value, np.ndarray):
            attributes = {
                "values": value.astype(np.int32),
            }
            self._format_transmitter_ids(value, attributes)
            value = self.add_data({"Transmitter ID": attributes})

        if not isinstance(value, (ReferencedData, IntegerData, type(None))):
            raise TypeError(
                "Input value for 'tx_id_property' should be of type uuid.UUID, "
                "ReferencedData, np.ndarray or None.)"
            )

        self._tx_id_property = value

        if self.type == "Receivers":
            self.edit_em_metadata({"Tx ID property": getattr(value, "uid", None)})
        else:
            self.edit_em_metadata({"Tx ID tx property": getattr(value, "uid", None)})

    @property
    def unit(self) -> float | None:
        """
        Default channel units for time or frequency defined on the child class.
        """
        return self.metadata["EM Dataset"].get("Unit")

    @unit.setter
    def unit(self, value: str):
        if self.default_units is not None:
            if value not in self.default_units:
                raise ValueError(f"Input 'unit' must be one of {self.default_units}")
            self.edit_em_metadata({"Unit": value})

    def validate_em_metadata(self, values: dict | np.ndarray | bytes | None) -> dict:
        """
        Validate and format the metadata structure for EM entities.

        :param values: Metadata dictionary.

        :return: Validated and formatted metadata dictionary.
        """
        if values is None:
            metadata = self.default_metadata

            if self.type is not None:
                metadata["EM Dataset"][self.type] = self.uid

            values = metadata

        if isinstance(values, np.ndarray):
            values = values[0]

        if isinstance(values, bytes):
            values = str_json_to_dict(values)

        if not isinstance(values, dict):
            raise TypeError("'metadata' must be of type 'dict'")

        if "EM Dataset" not in values:
            values = {"EM Dataset": values}

        missing_keys = []
        for key in self.default_metadata["EM Dataset"]:
            if key not in values["EM Dataset"]:
                missing_keys += [key]

        if missing_keys:
            raise KeyError(
                f"'{missing_keys}' argument(s) missing from the input metadata."
            )

        for key, value in values["EM Dataset"].items():
            values["EM Dataset"][key] = str2uuid(value)

        if "Property groups" in values["EM Dataset"]:
            prop_groups = []
            for value in values["EM Dataset"]["Property groups"]:
                prop_group = self.get_property_group(str2uuid(value))[0]

                if isinstance(prop_group, PropertyGroup):
                    prop_groups.append(prop_group.name)

            values["EM Dataset"]["Property groups"] = prop_groups

        return values

    def _edit_validate_property_groups(
        self, values: PropertyGroup | list[PropertyGroup] | list[str] | None
    ):
        """
        Add or append property groups to the metadata.

        :param value:
        """
        if not values:
            self.metadata["EM Dataset"]["Property groups"] = []
            return

        if not isinstance(values, list):
            values = [values]

        groups = (
            {group.name: group for group in self.property_groups}
            if self.property_groups
            else {}
        )

        for value in values:
            if self.property_groups is None:
                continue

            if not isinstance(value, (PropertyGroup, str)):
                raise TypeError(
                    "Input value for 'Property groups' must be a PropertyGroup or "
                    "name of an existing PropertyGroup."
                )

            if not (value in groups or value in groups.values()):
                raise ValueError("Property group must be an existing PropertyGroup.")

            if isinstance(value, str):
                value = groups[value]

            if value.properties is not None and len(value.properties) != len(
                self.channels
            ):
                raise ValueError(
                    f"Number of properties in group '{value.name}' "
                    + "differ from the number of 'channels'."
                )

            if value.name not in self.metadata["EM Dataset"]["Property groups"]:
                self.metadata["EM Dataset"]["Property groups"].append(value.name)

    def _fetch_transmitter_id(self) -> ReferencedData | IntegerData | None:
        """
        Utility method to retrieve the transmitter ID property, either from
        metadata or from list of children Data.
        """
        if self.type == "Receivers":
            uid = self.metadata["EM Dataset"].get("Tx ID property", None)
        else:
            uid = self.metadata["EM Dataset"].get("Tx ID tx property", None)

        if uid is not None:
            tx_id_property = self.get_entity(uid)[0]
        else:
            tx_id_property = self.get_entity("Transmitter ID")[0]
            # Reset the metadata

        if not isinstance(tx_id_property, ReferencedData | IntegerData | type(None)):
            raise TypeError(
                "Transmitter ID property must be of type ReferencedData or IntegerData."
            )

        return tx_id_property

    def _format_transmitter_ids(self, _, attributes):
        if self.complement is not None and self.complement.tx_id_property is not None:
            attributes["entity_type"] = self.complement.tx_id_property.entity_type

    def _get_complement_mask(
        self, mask: np.ndarray, complement: BaseEMSurvey
    ) -> np.ndarray:
        """
        Get the complement mask based on the input mask.

        :param mask: Mask on vertices or cells.
        :param complement: Complement entity.
        """
        if (
            self.tx_id_property is not None
            and complement.tx_id_property is not None
            and complement.tx_id_property.values is not None
            and mask is not None
        ):
            max_id = np.max(
                [
                    complement.tx_id_property.values.max(),
                    self.tx_id_property.values.max(),
                ]
            )
            ids_mask = np.zeros(max_id + 1, dtype=bool)
            ids_mask[self.tx_id_property.values] = True
            mask = ids_mask[complement.tx_id_property.values]

        return mask


class MovingLoopGroundEMSurvey(BaseEMSurvey, Curve, ABC):
    """
    A base class for moving loop ground EM surveys defined by a single transmitter
    loop radius.
    """

    __INPUT_TYPE = ["Rx"]
    _TYPE_UID: uuid.UUID | None = None

    @property
    def base_receiver_type(self):
        return Curve

    @property
    def base_transmitter_type(self):
        return Curve

    @property
    def default_input_types(self) -> list[str]:
        """Choice of survey creation types."""
        return self.__INPUT_TYPE

    @property
    def loop_radius(self) -> float | None:
        """Transmitter loop radius"""
        return self.metadata["EM Dataset"].get("Loop radius", None)

    @loop_radius.setter
    def loop_radius(self, value: float | None):
        if not isinstance(value, (float, type(None))):
            raise TypeError("Input 'loop_radius' must be of type 'float'")
        self.edit_em_metadata({"Loop radius": value})


class LargeLoopGroundEMSurvey(BaseEMSurvey, Curve, ABC):
    """
    Base class for large loop ground EM surveys defined by wire paths and transmitter ids.
    """

    __INPUT_TYPE = ["Tx and Rx"]
    _TYPE_UID: uuid.UUID | None = None

    @property
    def base_receiver_type(self):
        return Curve

    @property
    def base_transmitter_type(self):
        return Curve

    @property
    def default_input_types(self) -> list[str]:
        """Choice of survey creation types."""
        return self.__INPUT_TYPE

    def _format_transmitter_ids(self, values, attributes):
        if self.complement is not None and self.complement.tx_id_property is not None:
            attributes["entity_type"] = self.complement.tx_id_property.entity_type
        else:
            value_map = {
                ind: f"Loop {ind}" for ind in np.unique(values.astype(np.int32))
            }
            value_map[0] = "Unknown"
            attributes.update(
                {
                    "primitive_type": "REFERENCED",
                    "value_map": value_map,
                    "association": "VERTEX",
                }
            )

    def _renumber_reference_ids(self):
        """
        Re-number the value_map for tx_id_property to remain
        """
        if (
            self.tx_id_property is None
            or self.complement is None
            or self.complement.tx_id_property is None
        ):
            raise AttributeError(
                "The 'tx_id_property' must be set on both the object and its complement."
            )
        value_map = {
            val: ind
            for ind, val in enumerate(
                np.r_[0, np.unique(self.transmitters.tx_id_property.values)]
            )
        }
        new_map = self.complement.tx_id_property.entity_type.validate_value_map(
            {
                val: dict(self.transmitters.tx_id_property.value_map.map)[ind]
                for ind, val in value_map.items()
            }
        )
        self.complement.tx_id_property.values = np.asarray(
            [value_map[val] for val in self.complement.tx_id_property.values]
        )
        self.complement.tx_id_property.entity_type.value_map = new_map
        self.tx_id_property.values = np.asarray(
            [value_map[val] for val in self.tx_id_property.values]
        )
        self.tx_id_property.entity_type.value_map = new_map


class AirborneEMSurvey(BaseEMSurvey, Curve, ABC):
    __INPUT_TYPE = ["Rx", "Tx", "Tx and Rx"]
    _PROPERTY_MAP = {
        "crossline_offset": "Crossline offset",
        "inline_offset": "Inline offset",
        "pitch": "Pitch",
        "roll": "Roll",
        "vertical_offset": "Vertical offset",
        "yaw": "Yaw",
    }
    _TYPE_UID: uuid.UUID | None = None

    @property
    def crossline_offset(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the crossline offset between receiver and transmitter.
        """
        return self.fetch_metadata("crossline_offset")

    @crossline_offset.setter
    def crossline_offset(self, value: float | uuid.UUID | None):
        self.set_metadata("crossline_offset", value)

    @property
    def default_input_types(self) -> list[str]:
        """Choice of survey creation types."""
        return self.__INPUT_TYPE

    def fetch_metadata(self, key: str) -> float | uuid.UUID | None:
        """
        Fetch entry from the metadata.
        """
        field = self._PROPERTY_MAP.get(key, "")
        if field + " value" in self.metadata["EM Dataset"]:
            return self.metadata["EM Dataset"][field + " value"]
        if field + " property" in self.metadata["EM Dataset"]:
            return self.metadata["EM Dataset"][field + " property"]
        return None

    def set_metadata(self, key: str, value: float | uuid.UUID | None):
        if key not in self._PROPERTY_MAP:
            raise ValueError(f"No property map found for key metadata '{key}'.")

        field = self._PROPERTY_MAP[key]
        if isinstance(value, float):
            self.edit_em_metadata({field + " value": value, field + " property": None})
        elif isinstance(value, uuid.UUID):
            self.edit_em_metadata({field + " value": None, field + " property": value})
        elif value is None:
            self.edit_em_metadata({field + " value": None, field + " property": None})
        else:
            raise TypeError(
                f"Input '{key}' must be one of type float, uuid.UUID or None"
            )

    @property
    def inline_offset(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the inline offset between receiver and transmitter.
        """
        return self.fetch_metadata("inline_offset")

    @inline_offset.setter
    def inline_offset(self, value: float | uuid.UUID):
        self.set_metadata("inline_offset", value)

    @property
    def loop_radius(self) -> float | None:
        """Transmitter loop radius"""
        return self.metadata["EM Dataset"].get("Loop radius", None)

    @loop_radius.setter
    def loop_radius(self, value: float | None):
        if not isinstance(value, (float, type(None))):
            raise TypeError("Input 'loop_radius' must be of type 'float'")
        self.edit_em_metadata({"Loop radius": value})

    @property
    def pitch(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the pitch angle of the transmitter loop.
        """
        return self.fetch_metadata("pitch")

    @pitch.setter
    def pitch(self, value: float | uuid.UUID | None):
        self.set_metadata("pitch", value)

    @property
    def relative_to_bearing(self) -> bool | None:
        """Data relative_to_bearing"""
        return self.metadata["EM Dataset"].get("Angles relative to bearing", None)

    @relative_to_bearing.setter
    def relative_to_bearing(self, value: bool | None):
        if not isinstance(value, (bool, type(None))):
            raise TypeError("Input 'relative_to_bearing' must be one of type 'bool'")
        self.edit_em_metadata({"Angles relative to bearing": value})

    @property
    def roll(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the roll angle of the transmitter loop.
        """
        return self.fetch_metadata("roll")

    @roll.setter
    def roll(self, value: float | uuid.UUID | None):
        self.set_metadata("roll", value)

    @property
    def vertical_offset(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the vertical offset between receiver and transmitter.
        """
        return self.fetch_metadata("vertical_offset")

    @vertical_offset.setter
    def vertical_offset(self, value: float | uuid.UUID | None):
        self.set_metadata("vertical_offset", value)

    @property
    def yaw(self) -> float | uuid.UUID | None:
        """
        Numeric value or property UUID for the yaw angle of the transmitter loop.
        """
        return self.fetch_metadata("yaw")

    @yaw.setter
    def yaw(self, value: float | uuid.UUID):
        self.set_metadata("yaw", value)


class FEMSurvey(BaseEMSurvey, ABC):
    __UNITS = __UNITS = [
        "Hertz (Hz)",
        "KiloHertz (kHz)",
        "MegaHertz (MHz)",
        "Gigahertz (GHz)",
    ]

    @property
    def default_units(self) -> list[str]:
        """
        Accepted frequency units.

        Must be one of "Hertz (Hz)", "KiloHertz (kHz)", "MegaHertz (MHz)", or
        "Gigahertz (GHz)",

        :returns: List of acceptable units for frequency domain channels.
        """
        return self.__UNITS


class TEMSurvey(BaseEMSurvey, ABC):
    __UNITS = [
        "Seconds (s)",
        "Milliseconds (ms)",
        "Microseconds (us)",
        "Nanoseconds (ns)",
    ]

    @property
    def default_units(self) -> list[str]:
        """
        Accepted time units.

        Must be one of "Seconds (s)", "Milliseconds (ms)", "Microseconds (us)"
        or "Nanoseconds (ns)"

        :returns: List of acceptable units for time domain channels.
        """
        return self.__UNITS

    @property
    def timing_mark(self) -> float | None:
        """
        Timing mark from the beginning of the discrete :attr:`waveform`.
        Generally used as the reference (time=0.0) for the provided
        (-) on-time an (+) off-time :attr:`channels`.
        """
        if (
            "Waveform" in self.metadata["EM Dataset"]
            and "Timing mark" in self.metadata["EM Dataset"]["Waveform"]
        ):
            timing_mark = self.metadata["EM Dataset"]["Waveform"]["Timing mark"]
            return timing_mark

        return None

    @timing_mark.setter
    def timing_mark(self, timing_mark: float | None):
        if not isinstance(timing_mark, (float, type(None))):
            raise ValueError("Input timing_mark must be a float or None.")

        if self.waveform is not None:
            value = self.metadata["EM Dataset"]["Waveform"]
        else:
            value = {}

        if timing_mark is None and "Timing mark" in value:
            del value["Timing mark"]
        else:
            value["Timing mark"] = timing_mark

        self.edit_em_metadata({"Waveform": value})

    @property
    def waveform(self) -> np.ndarray | None:
        """
        Discrete waveform of the TEM source provided as
        :obj:`numpy.array` of type :obj:`float`, shape(n, 2)

        .. code-block:: python

            waveform = [
                [time_1, current_1],
                [time_2, current_2],
                ...
            ]

        """
        if (
            "Waveform" in self.metadata["EM Dataset"]
            and "Discretization" in self.metadata["EM Dataset"]["Waveform"]
        ):
            waveform = np.vstack(
                [
                    [row["time"], row["current"]]
                    for row in self.metadata["EM Dataset"]["Waveform"]["Discretization"]
                ]
            )
            return waveform
        return None

    @waveform.setter
    def waveform(self, waveform: np.ndarray | None):
        if not isinstance(waveform, (np.ndarray, type(None))):
            raise TypeError("Input waveform must be a numpy.ndarray or None.")

        if self.timing_mark is not None:
            value = self.metadata["EM Dataset"]["Waveform"]
        else:
            value = {"Timing mark": 0.0}

        if isinstance(waveform, np.ndarray):
            if waveform.ndim != 2 or waveform.shape[1] != 2:
                raise ValueError(
                    "Input waveform must be a numpy.ndarray of shape (*, 2)."
                )

            value["Discretization"] = [
                {"current": row[1], "time": row[0]} for row in waveform
            ]

        self.edit_em_metadata({"Waveform": value})

    @property
    def waveform_parameters(self) -> dict | None:
        """Access the waveform parameters stored as a dictionary."""
        waveform = self.get_data("_waveform_parameters")[0]

        if waveform is not None:
            return json.loads(waveform.values)

        return None



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

import numpy as np
from PIL import Image

from ... import objects
from ...data import Data
from ...shared import FLOAT_NDV
from ...shared.utils import PILLOW_ARGUMENTS, xy_rotation_matrix, yz_rotation_matrix
from .base import CellObjectConversion


if TYPE_CHECKING:
    from ...objects import GeoImage, Grid2D


class Grid2DConversion(CellObjectConversion):
    """
    Convert a :obj:'geoh5py.objects.grid2d.Grid2D' object
    to a georeferenced :obj:'geoh5py.objects.geo_image.GeoImage' object.
    """

    # convert to geoimage

    @staticmethod
    def grid_to_tag(input_entity: Grid2D) -> dict:
        """
        Compute the tag dictionary of the :obj:'geoh5py.objects.geo_image.Grid2D'
        as required by the :obj:'geoh5py.objects.geo_image.GeoImage' object.
        :param input_entity: the Grid2D object to convert.
        :return tag: the tag dictionary.
        """
        if not isinstance(input_entity.origin, np.ndarray):
            raise AttributeError("The Grid2D has no origin.")

        if input_entity.v_cell_size is None or input_entity.u_cell_size is None:
            raise AttributeError("The Grid2D has no cell size information.")

        if input_entity.u_count is None or input_entity.v_count is None:
            raise AttributeError("The Grid2D has no number of cells.")

        u_origin, v_origin, z_origin = input_entity.origin.tolist()
        v_oposite = v_origin + input_entity.v_cell_size * input_entity.v_count

        tag = {
            256: (input_entity.u_count,),
            257: (input_entity.v_count,),
            33550: (
                input_entity.u_cell_size,
                input_entity.v_cell_size,
                0.0,
            ),
            33922: (
                0.0,
                0.0,
                0.0,
                u_origin,
                v_oposite,
                z_origin,
            ),
        }

        return tag

    @staticmethod
    def data_to_pil_format(
        input_entity: Grid2D, data: np.ndarray, normalize: bool = True
    ) -> np.ndarray:
        """
        Convert a numpy array with a format compatible with :obj:`PIL.Image` object.
        :param input_entity: the Grid2D object to convert.
        :param data: the data to convert.
        :param normalize: if True, the data will be normalized between 0 and 255.
        :return: the data formatted with the right shape,
        between 0 and 255, as uint8.
        """
        # reshape them
        data = data.reshape(input_entity.v_count, input_entity.u_count)[::-1]

        # remove nan values
        data = np.where(data == FLOAT_NDV, np.nan, data)

        # normalize them
        if normalize:
            min_, max_ = np.nanmin(data), np.nanmax(data)
            data = (data - min_) / (max_ - min_)
            data *= 255

        return data.astype(np.uint8)

    @staticmethod
    def key_to_data(input_entity: Grid2D, key: str | int | UUID | Data) -> np.ndarray:
        """
        Extract the data from the entity in :obj:'np.array' format;
        The data can be of type: ':obj:str', ':obj:int', ':obj:UUID', or ':obj:Data'.
        :param input_entity: the Grid2D object to convert.
        :param key: the key of the data to extract.
        :return: an np. array containing the data.
        """
        # get the values
        if isinstance(key, str):
            data = input_entity.get_data(key)
        elif isinstance(key, int):
            if key > len(input_entity.get_entity_list()):
                raise IndexError(
                    "'int' values pass as key can't be larger than number of data,",
                    f"data number: {len(input_entity.get_entity_list())}, key: {key}",
                )
            key_ = input_entity.get_entity_list()[key]
            data = input_entity.get_entity(key_)  # type: ignore
        elif isinstance(key, UUID):
            data = input_entity.get_entity(key)  # type: ignore
        elif isinstance(key, Data):
            data = key  # type: ignore
        else:
            raise TypeError(
                "The dtype of the keys must be :",
                "'str', 'int', 'Data', 'UUID',",
                f"but you entered a {type(key)}",
            )

        # verify if data exists
        if isinstance(data, list):
            if len(data) != 1:
                raise KeyError(
                    f"The key '{key}' you entered does not exists;"
                    f"Valid keys are: {input_entity.get_data_list()}."
                )
            data = data[0]  # type: ignore

        return data.values  # type: ignore

    @staticmethod
    def data_from_keys(
        input_entity: Grid2D,
        keys: list | str | int | UUID | Data,
        normalize: bool = True,
    ) -> np.ndarray:
        """
        Take a list of (or a unique) key to extract from the object,
        and create a :obj:'np.array' with those data.
        :param input_entity: the Grid2D object to convert.
        :param normalize: if True, the data will be normalized between 0 and 255.
        :param keys: the list of the data to extract.
        :return data: the data extracted from the object.
        """

        # if unique key transform to list
        if isinstance(keys, (str, int, UUID, Data)):
            keys = [keys]

        # prepare the image
        if isinstance(keys, list):
            data = np.empty((input_entity.v_count, input_entity.u_count, 0)).astype(
                np.uint8
            )

            for key in keys:
                data_temp = Grid2DConversion.key_to_data(input_entity, key)

                data_temp = Grid2DConversion.data_to_pil_format(
                    input_entity, data_temp, normalize
                )

                data = np.dstack((data, data_temp))

            return data

        raise TypeError(
            "The keys must be pass as a list",
            f"but you entered a {type(keys)} ",
        )

    @staticmethod
    def convert_to_pillow(data: np.ndarray, mode: str | None = None) -> Image:
        """
        Convert the data from :obj:'np.array' to :obj:'PIL.Image' format.
        """
        if not isinstance(data, np.ndarray):
            raise AttributeError("No data is selected.")

        if mode is None:
            if data.shape[-1] == 1:
                mode = "L"
                data = data[:, :, 0]

            if data.shape[-1] == 3:
                mode = "RGB"
            elif data.shape[-1] == 4:
                mode = "RGBA"

            if mode is None:
                raise IndexError("Only 1, 3, or 4 layers can be selected")

        if mode not in PILLOW_ARGUMENTS:
            raise NotImplementedError(f"The mode {mode} is not actually supported.")

        return Image.fromarray(data, mode=mode)

    @staticmethod
    def compute_vertices(input_entity: Grid2D) -> np.ndarray:
        """
        Compute the vertices of the geoimage to create based on the
        properties of the grid and its angle.
        :param input_entity: The grid2D object to convert.
        :return: A numpy array of 4 points in x,y,z.
        """
        # get the length of the x and y axis
        if (
            input_entity.u_count is None
            or input_entity.v_count is None
            or input_entity.u_cell_size is None
            or input_entity.v_cell_size is None
        ):
            raise AttributeError(
                "The grid2D object must have the following properties"
                ": u_count, v_count, u_cell_size, v_cell_size"
            )

        length_u = input_entity.u_count * input_entity.u_cell_size
        length_v = input_entity.v_count * input_entity.v_cell_size

        # Define the corners
        corners = np.array(
            [[0, length_v, 0], [length_u, length_v, 0], [length_u, 0, 0], [0, 0, 0]]
        )

        # rotate by the dip
        dip_matrix = yz_rotation_matrix(np.deg2rad(input_entity.dip))
        dipped_corners = dip_matrix @ corners.T

        # rotate by the rotation
        rotation_matrix = xy_rotation_matrix(np.deg2rad(input_entity.rotation))
        rotated_corners = rotation_matrix @ dipped_corners

        # add the origin
        shifted_corners = rotated_corners.T + np.array(input_entity.origin.tolist())

        return shifted_corners

    @staticmethod
    def to_geoimage(
        input_entity: Grid2D,
        keys: list | str | int | UUID | Data,
        mode: str | None = None,
        normalize: bool = True,
        **geoimage_kwargs,
    ) -> GeoImage:
        """
        Convert the object to a :obj:'GeoImage' object.
        :param input_entity: the Grid2D object to convert.
        :param keys: the data to extract.
        :param mode: the mode of the image to create.
        :param normalize: if True, the data will be normalized between 0 and 255.
        :param geoimage_kwargs: the kwargs to pass to the :obj:'GeoImage' object.
        """
        workspace, geoimage_kwargs = Grid2DConversion.validate_workspace(
            input_entity, **geoimage_kwargs
        )

        geoimage_kwargs = Grid2DConversion.verify_kwargs(
            input_entity, **geoimage_kwargs
        )

        # get the tag of the data
        geoimage_kwargs["tag"] = Grid2DConversion.grid_to_tag(input_entity)

        # get the data
        data = Grid2DConversion.data_from_keys(input_entity, keys, normalize=normalize)

        # define the image
        geoimage_kwargs["image"] = Grid2DConversion.convert_to_pillow(data, mode=mode)

        # define the vertices
        geoimage_kwargs["vertices"] = Grid2DConversion.compute_vertices(input_entity)

        # create a geoimage
        output = objects.GeoImage.create(workspace, **geoimage_kwargs)

        return output



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=R0902, R0904

from __future__ import annotations

import re
import uuid
import warnings
from numbers import Real

import numpy as np

from ..data import Data, FloatData, NumericData
from ..shared.utils import (
    box_intersect,
    dip_azimuth_to_vector,
    mask_by_extent,
    merge_arrays,
)
from .points import Points


INFINITE_RADIUS = 99999.9
MINIMUM_DEPTH_INTERVAL = 1.0
MAXIMUM_DEPTH_INTERVAL = 50.0


class Drillhole(Points):
    """
    Drillhole object class defined by a collar and survey.

    :param collar: Coordinates of the drillhole.
    :param cost: Cost estimate of the drillhole.
    :param end_of_hole: End of drillhole in meters.
    :param planning: Status of the hole, defaults to 'Default'.
    :param surveys: Survey information provided as 'Depth', 'Azimuth', 'Dip'.
    :param vertices: Coordinates of the vertices.
    :param default_collocation_distance: Minimum collocation distance for matching depth on merge.
    """

    _TYPE_UID = uuid.UUID(
        fields=(0x7CAEBF0E, 0xD16E, 0x11E3, 0xBC, 0x69, 0xE4632694AA37)
    )
    __SURVEY_DTYPE = np.dtype([("Depth", "<f4"), ("Azimuth", "<f4"), ("Dip", "<f4")])
    __COLLAR_DTYPE = np.dtype([("x", float), ("y", float), ("z", float)])
    _attribute_map = Points._attribute_map.copy()
    _attribute_map.update(
        {
            "Cost": "cost",
            "Collar": "collar",
            "Planning": "planning",
            "End of hole": "end_of_hole",
        }
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        collar: np.ndarray | list | None = None,
        cost: float = 0.0,
        end_of_hole: float | None = None,
        planning: str = "Default",
        surveys: np.ndarray | list | tuple | None = None,
        vertices: np.ndarray | None = None,
        default_collocation_distance: float = 1e-2,
        **kwargs,
    ):
        self._cells: np.ndarray | None = None
        self._depths: FloatData | None = None
        self._trace: np.ndarray | None = None
        self._trace_depth: np.ndarray | None = None
        self._locations = None
        self._surveys: np.ndarray | None = None
        self._intervals: dict | None = None

        super().__init__(
            vertices=(0.0, 0.0, 0.0) if vertices is None else vertices, **kwargs
        )

        if vertices is None:
            self._vertices = None

        self.collar = collar
        self.cost = cost
        self.default_collocation_distance = default_collocation_distance
        self.end_of_hole = end_of_hole
        self.planning = planning

        if surveys is not None:
            self.surveys = surveys

    @property
    def cells(self) -> np.ndarray | None:
        r"""
        :obj:`numpy.ndarray` of :obj:`int`, shape (\*, 2):
        Array of indices defining segments connecting vertices.
        """
        if getattr(self, "_cells", None) is None:
            if self.on_file:
                self._cells = self.workspace.fetch_array_attribute(self)

        return self._cells

    @cells.setter
    def cells(self, indices):
        if indices.dtype != "uint32":
            raise TypeError("Indices array must be of type 'uint32'")
        self._cells = indices
        self.workspace.update_attribute(self, "cells")

    @property
    def collar(self) -> np.ndarray:
        """
        Coordinates of the collar, shape(1, 3)
        """
        return self._collar

    @collar.setter
    def collar(self, value: list | np.ndarray | None):
        if value is None:
            warnings.warn(
                "No 'collar' provided. Using (0, 0, 0) as default point at the origin.",
                UserWarning,
            )
            value = (0.0, 0.0, 0.0)

        if isinstance(value, np.ndarray):
            value = value.tolist()

        if isinstance(value, str):
            value = [float(n) for n in re.findall(r"-?\d+\.\d+", value)]

        if len(value) != 3:
            raise ValueError("Origin must be a list or numpy array of len (3,).")

        value = np.asarray(tuple(value), dtype=self.__COLLAR_DTYPE)
        self._collar = value
        self._locations = None
        self._intervals = None

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

            if self.trace is not None:
                self._trace = None
                self._trace_depth = None
                self.workspace.update_attribute(self, "trace")
                self.workspace.update_attribute(self, "trace_depth")

    @property
    def cost(self) -> float:
        """
        :obj:`float`: Cost estimate of the drillhole
        """
        return self._cost

    @cost.setter
    def cost(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError(f"Provided cost value must be of type {float} or int.")
        self._cost = float(value)

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def end_of_hole(self) -> float | None:
        """
        End of drillhole in meters
        """
        return self._end_of_hole

    @end_of_hole.setter
    def end_of_hole(self, value: Real | None):
        if not isinstance(value, (int, float, type(None))):
            raise TypeError(f"Provided end_of_hole value must be of type {int}")
        self._end_of_hole = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def extent(self) -> np.ndarray | None:
        """
        Geography bounding box of the object.

        :return: shape(2, 3) Bounding box defined by the bottom South-West and
            top North-East coordinates.
        """
        if self.collar is not None:
            return (
                np.repeat(
                    np.r_[[self.collar["x"], self.collar["y"], self.collar["z"]]], 2
                )
                .reshape((-1, 2))
                .T
            )

        return None

    @property
    def intervals(self):
        """
        Densified and computed interval information
        """
        if self._intervals is None:
            self._intervals = self.compute_intervals(
                self.surveys, self.collar.tolist(), self.end_of_hole
            )
        return self._intervals

    @property
    def locations(self) -> np.ndarray | None:
        """
        Lookup array of the well path in x, y, z coordinates.
        """
        if getattr(self, "_locations", None) is None and self.collar is not None:
            self._locations = self.depths_to_xyz(self.intervals, self.surveys[:, 0])

        return self._locations

    def mask_by_extent(
        self, extent: np.ndarray, inverse: bool = False
    ) -> np.ndarray | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.

        Uses the collar location only.
        """
        if self.extent is None or not box_intersect(self.extent, extent):
            return None

        if self.collar is not None:
            return mask_by_extent(
                np.c_[[self.collar["x"], self.collar["y"], self.collar["z"]]].T,
                extent,
                inverse=inverse,
            )

        return None

    @property
    def n_cells(self) -> int | None:
        """
        Number of cells.
        """
        if self.cells is not None:
            return self.cells.shape[0]
        return None

    @property
    def planning(self) -> str:
        """
        Status of the hole on of "Default", "Ongoing", "Planned", "Completed" or "No status"
        """
        return self._planning

    @planning.setter
    def planning(self, value: str):
        choices = ["Default", "Ongoing", "Planned", "Completed", "No status"]
        if value not in choices:
            raise ValueError(f"Provided planning value must be one of {choices}")
        self._planning = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def surveys(self) -> np.ndarray:
        """
        Coordinates of the surveys.
        """
        if self._surveys is None and self.on_file:
            self._surveys = self.workspace.fetch_array_attribute(self, "surveys")

        if self._surveys is None:
            surveys = np.c_[0, 0, -90]

        else:
            surveys = np.vstack(
                [
                    self._surveys["Depth"],
                    self._surveys["Azimuth"],
                    self._surveys["Dip"],
                ]
            ).T

        return surveys.astype(float)

    @surveys.setter
    def surveys(self, array: np.ndarray | list | tuple):
        if not isinstance(array, (np.ndarray, list, tuple)):
            raise TypeError(
                "Input 'surveys' must be of type 'numpy.ndarray' or 'list'."
            )

        self._surveys = self.format_survey_values(array)
        self.end_of_hole = float(self._surveys["Depth"][-1])

        if self.on_file:
            self.workspace.update_attribute(self, "surveys")

            if self.trace is not None:
                self._trace = None
                self.workspace.update_attribute(self, "trace")

        self._locations = None
        self._intervals = None

    def format_survey_values(self, values: list | tuple | np.ndarray) -> np.ndarray:
        """
        Reformat the survey values as structured array with the right shape.
        """
        if isinstance(values, (list, tuple)):
            values = np.array(values, ndmin=2)

        if np.issubdtype(values.dtype, np.number):
            if values.shape[1] != 3:
                raise ValueError("'surveys' requires an ndarray of shape (*, 3)")

            array_values = np.asarray(
                np.core.records.fromarrays(values.T, dtype=self.__SURVEY_DTYPE)
            )
        else:
            array_values = values

        if array_values.dtype.descr[:3] != self.__SURVEY_DTYPE.descr:
            raise ValueError(
                f"Array of 'survey' must be of dtype = {self.__SURVEY_DTYPE}"
            )
        return array_values

    @property
    def default_collocation_distance(self):
        """
        Minimum collocation distance for matching depth on merge
        """
        return self._default_collocation_distance

    @default_collocation_distance.setter
    def default_collocation_distance(self, tol):
        if not tol > 0:
            raise ValueError("Tolerance value should be >0.")

        self._default_collocation_distance = tol
        self.workspace.update_attribute(self, "attributes")

    @property
    def trace(self) -> np.ndarray | None:
        """
        :obj:`numpy.array`: Drillhole trace defining the path in 3D
        """
        if self._trace is None and self.on_file:
            self._trace = self.workspace.fetch_array_attribute(self, "trace")

        if self._trace is not None:
            return self._trace.view("<f8").reshape((-1, 3))

        return None

    @property
    def trace_depth(self) -> np.ndarray | None:
        """
        :obj:`numpy.array`: Drillhole trace depth from top to bottom
        """
        if getattr(self, "_trace_depth", None) is None and self.trace is not None:
            trace = self.trace
            self._trace_depth = trace[0, 2] - trace[:, 2]

        return self._trace_depth

    @property
    def from_(self):
        """
        Depth data corresponding to the tops of the interval values.
        """
        data_obj = self.get_data("FROM")
        if data_obj:
            return data_obj[0]

        return None

    @property
    def to_(self):
        """
        Depth data corresponding to the bottoms of the interval values.
        """
        data_obj = self.get_data("TO")
        if data_obj:
            return data_obj[0]

        return None

    @property
    def depths(self) -> FloatData | None:
        if self._depths is None:
            data_obj = self.get_data("DEPTH")
            if data_obj and isinstance(data_obj[0], FloatData):
                self.depths = data_obj[0]
        return self._depths

    @depths.setter
    def depths(self, value: FloatData | np.ndarray | None):
        if isinstance(value, np.ndarray):
            value = self.workspace.create_entity(
                Data,
                entity={
                    "parent": self,
                    "association": "VERTEX",
                    "name": "DEPTH",
                    "values": value,
                },
                entity_type={"primitive_type": "FLOAT"},
            )

        if isinstance(value, (FloatData, type(None))):
            self._depths = value
        else:
            raise ValueError(
                f"Input '_depth' property must be of type{FloatData} or None"
            )

    def desurvey(self, depths):
        """
        Function to return x, y, z coordinates from depth.
        """
        return self.depths_to_xyz(self.intervals, depths)

    def add_vertices(self, xyz):
        """
        Function to add vertices to the drillhole
        """
        indices = np.arange(xyz.shape[0])
        if self._vertices is not None:
            indices += self.n_vertices
            xyz = np.vstack([self.vertices, xyz])

        self._vertices = np.asarray(
            np.core.records.fromarrays(
                xyz.T.tolist(),
                dtype=[("x", "<f8"), ("y", "<f8"), ("z", "<f8")],
            )
        )
        self.workspace.update_attribute(self, "vertices")

        return indices.astype("uint32")

    def validate_interval_data(  # pylint: disable=too-many-locals
        self,
        from_to: np.ndarray | list,
        values: np.ndarray,
        collocation_distance: float = 1e-4,
    ) -> np.ndarray:
        """
        Compare new and current depth values, append new vertices if necessary and return
        an augmented values vector that matches the vertices indexing.

        :param from_to: Array of from-to values.
        :param values: Array of values.
        :param collocation_distance: Minimum collocation distance for matching.

        :return: Augmented values vector that matches the vertices indexing.
        """
        if isinstance(from_to, list):
            from_to = np.vstack(from_to)

        if from_to.shape[0] != len(values):
            raise ValueError(
                f"Mismatch between input 'from_to' shape{from_to.shape} "
                + f"and 'values' shape{values.shape}"
            )

        if from_to.ndim != 2 or from_to.shape[1] != 2:
            raise ValueError("The `from-to` values must have shape(*, 2).")

        if (self.from_ is None) and (self.to_ is None):
            uni_depth, inv_map = np.unique(from_to, return_inverse=True)
            self.cells = self.add_vertices(self.desurvey(uni_depth))[inv_map].reshape(
                (-1, 2)
            )
            self.workspace.create_entity(
                Data,
                entity={
                    "parent": self,
                    "association": "CELL",
                    "name": "FROM",
                    "values": from_to[:, 0],
                },
                entity_type={"primitive_type": "FLOAT"},
            )
            self.workspace.create_entity(
                Data,
                entity={
                    "parent": self,
                    "association": "CELL",
                    "name": "TO",
                    "values": from_to[:, 1],
                },
                entity_type={"primitive_type": "FLOAT"},
            )
        elif self.cells is not None and self.from_ is not None and self.to_ is not None:
            values = np.r_[values]
            out_vec = np.c_[self.from_.values, self.to_.values]
            dist_match = []
            for i, elem in enumerate(from_to):
                ind = np.where(
                    np.linalg.norm(elem - out_vec, axis=1) < collocation_distance
                )[0]
                if len(ind) > 0:
                    dist_match.append([ind[0], i])

            cell_map: np.ndarray = np.asarray(dist_match, dtype=int)

            # Add vertices
            vert_new = np.ones_like(from_to, dtype="bool")
            if cell_map.ndim == 2:
                vert_new[cell_map[:, 1], :] = False
            ind_new = np.where(vert_new.flatten())[0]
            uni_new, inv_map = np.unique(
                from_to.flatten()[ind_new], return_inverse=True
            )

            # check if its text data, and defined nan array if so
            if values.dtype.kind in ["U", "S"]:
                nan_values = np.array([""] * self.n_cells)  # type: ignore
            else:
                nan_values = np.ones(self.n_cells) * np.nan

            # Append values
            values = merge_arrays(
                nan_values,
                values,
                replace="B->A",
                mapping=cell_map,
            )
            self.cells = np.r_[
                self.cells,
                self.add_vertices(self.desurvey(uni_new))[inv_map]
                .reshape((-1, 2))
                .astype("uint32"),
            ]
            self.from_.values = merge_arrays(
                self.from_.values, from_to[:, 0], mapping=cell_map
            )
            self.to_.values = merge_arrays(
                self.to_.values, from_to[:, 1], mapping=cell_map
            )

        return values

    def validate_depth_data(
        self,
        depth: np.ndarray,
        values: np.ndarray,
        collocation_distance=1e-4,
    ) -> np.ndarray:
        """
        Compare new and current depth values. Append new vertices if necessary and return
        an augmented values vector that matches the vertices indexing.

        :param depth: Array of depth values.
        :param values: Array of values.
        :param collocation_distance: Minimum collocation distance for matching.

        :return: Augmented values vector that matches the vertices indexing.
        """
        if depth.shape != values.shape:
            raise ValueError(
                f"Mismatch between input 'depth' shape{depth.shape} "
                + f"and 'values' shape{values.shape}"
            )

        input_values = np.r_[values]

        if self.depths is None:
            self.add_vertices(self.desurvey(depth))
            self.depths = np.r_[
                np.ones(self.n_vertices - depth.shape[0]) * np.nan, depth
            ]
            values = np.r_[
                np.ones(self.n_vertices - input_values.shape[0]) * np.nan, input_values
            ]
        else:
            depths, indices = merge_arrays(
                self.depths.values,
                depth,
                return_mapping=True,
                collocation_distance=collocation_distance,
            )
            values = merge_arrays(
                np.ones(self.n_vertices) * np.nan,
                input_values,
                replace="B->A",
                mapping=indices,
            )
            self.add_vertices(self.desurvey(np.delete(depth, indices[:, 1])))
            self.depths.values = depths

        return values

    def validate_association(
        self, attributes: dict, property_group=None, collocation_distance=None, **_
    ) -> tuple:
        """
        Validate input drillhole data attributes.

        :param attributes: Dictionary of data attributes.
        :param property_group: Input property group to validate against.
        :param collocation_distance: Minimum collocation distance for matching.
        """
        if collocation_distance is None:
            collocation_distance = attributes.get(
                "collocation_distance", self.default_collocation_distance
            )

        if collocation_distance < 0:
            raise UserWarning("Input depth 'collocation_distance' must be >0.")

        if "depth" not in attributes and "from-to" not in attributes:
            if "association" not in attributes or attributes["association"] != "OBJECT":
                raise ValueError(
                    "Input data dictionary must contain a key/value pair of depth data "
                    "or contain an 'OBJECT' association. Valid depth keys are 'depth' "
                    "and 'from-to'."
                )

        if attributes["name"] in self.get_data_list():
            raise ValueError(
                f"Data with name '{attributes['name']}' already present "
                f"on the drillhole '{self.name}'. "
                "Consider changing the values or renaming."
            )

        if "depth" in attributes.keys():
            attributes["association"] = "VERTEX"
            attributes["values"] = self.validate_depth_data(
                attributes["depth"],
                attributes["values"],
                collocation_distance=collocation_distance,
            )
            del attributes["depth"]

        if "from-to" in attributes.keys():
            attributes["association"] = "CELL"
            attributes["values"] = self.validate_interval_data(
                attributes["from-to"],
                attributes["values"],
                collocation_distance=collocation_distance,
            )
            del attributes["from-to"]

        return attributes, property_group

    def validate_mask(self, mask: np.ndarray | None) -> np.ndarray | None:
        """
        Validate mask array.

        Always return None for drillholes such that all data is preserved.

        :param mask: None
        """
        return None

    @property
    def vertices(self) -> np.ndarray | None:
        """
        :obj:`~geoh5py.objects.object_base.ObjectBase.vertices`
        """
        if self._vertices is None and self.on_file:
            self._vertices = self.workspace.fetch_array_attribute(self, "vertices")

        if self._vertices is not None:
            return self._vertices.view("<f8").reshape((-1, 3))

        return None

    @vertices.setter
    def vertices(self, vertices: np.ndarray | list | tuple):
        xyz = self.validate_vertices(vertices)
        if self._vertices is not None and self._vertices.shape != xyz.shape:
            raise ValueError(
                "New vertices array must have the same shape as the current vertices array."
            )
        self._vertices = xyz

        self.workspace.update_attribute(self, "vertices")

    def post_processing(self):
        """
        Read the 'DEPTH' data and sort all Data.values if needed
        """
        if self.get_data("DEPTH"):
            data_obj = self.get_data("DEPTH")[0]

            depths = data_obj.values
            if isinstance(data_obj, NumericData):
                depths = data_obj.validate_values(depths)

            if not np.all(np.diff(depths) >= 0):
                sort_ind = np.argsort(depths)

                for child in self.children:
                    if (
                        isinstance(child, NumericData)
                        and getattr(child.association, "name", None) == "VERTEX"
                    ):
                        child.values = child.validate_values(child.values)[sort_ind]

                if self.vertices is not None:
                    self._vertices = self.vertices[sort_ind, :]
                    self.workspace.update_attribute(self, "vertices")

                if self.cells is not None:
                    key_map = np.argsort(sort_ind)[self.cells.flatten()]
                    self._cells = key_map.reshape((-1, 2)).astype("uint32")
                    self.workspace.update_attribute(self, "cells")

    @staticmethod
    def densify_survey_values(
        survey: np.ndarray, end_of_hole: float | None
    ) -> np.ndarray:
        """
        Validate the survey values for desurveying.

        - Repeat first survey point at surface for de-survey interpolation
        - Repeat last survey point at end of hole for de-survey interpolation
        - Densify along intervals if greater than 50 m

        :param survey: Array of survey points with columns 'Depth', 'Azimuth', 'Dip'.
        :param end_of_hole: End of the drillhole in meters.

        :return: Augmented array of survey points with columns 'Depth', 'Azimuth', 'Dip'.
        """
        full_survey = survey.copy()

        # Repeat first survey point at surface for de-survey interpolation
        if survey[0, 0] != 0.0:
            full_survey = np.vstack([survey[0, :], full_survey])
            full_survey[0, 0] = 0.0

        if end_of_hole is not None and end_of_hole > full_survey[-1, 0]:
            last_row = full_survey[-1, :].copy()
            last_row[0] = end_of_hole
            full_survey = np.vstack([full_survey, last_row])

        # Densify along intervals if greater than 50 m
        deltas = np.diff(full_survey[:, 0])
        new_points = []
        for ii, delta in enumerate(deltas):
            if delta < MAXIMUM_DEPTH_INTERVAL:
                continue

            depths = np.arange(0, delta, MAXIMUM_DEPTH_INTERVAL)
            d_azm = (full_survey[ii + 1, 1] - full_survey[ii, 1]) / delta
            d_dip = (full_survey[ii + 1, 2] - full_survey[ii, 2]) / delta
            new_points.append(
                np.c_[
                    full_survey[ii, 0] + depths[1:],
                    full_survey[ii, 1] + d_azm * depths[1:],
                    full_survey[ii, 2] + d_dip * depths[1:],
                ]
            )

        if len(new_points) > 1:
            full_survey = np.vstack([full_survey, np.vstack(new_points)])

        return np.unique(full_survey, axis=0)

    @staticmethod
    def compute_intervals(survey: np.ndarray, collar, end_of_hole) -> dict:
        """
        Compute the intervals from survey and collar information.

        Stores a dictionary describing the arc circle between the survey points
        separated by a maximum depth interval of 50 m.

        Translated from C++ code.

        :param survey: Array of survey points with columns 'Depth', 'Azimuth', 'Dip'.
        :param collar: Collar location of the drillhole.
        :param end_of_hole: End of the drillhole in meters.
        """
        collar = np.asarray(collar).reshape((1, 3))
        full_survey = Drillhole.densify_survey_values(survey, end_of_hole)
        unit_vector = dip_azimuth_to_vector(full_survey[:, 2], full_survey[:, 1])

        if survey.shape[0] < 2:
            return {
                "depths": np.r_[0],
                "rad": np.r_[INFINITE_RADIUS],
                "tangential": np.zeros((1, 3)),
                "unit_vector": unit_vector,
                "locations": collar,
            }

        # Cross product between neighbours
        cross = np.cross(unit_vector[:-1, :], unit_vector[1:, :], axis=1)
        vr = np.linalg.norm(cross, axis=1)
        dot = np.sum(unit_vector[:-1, :] * unit_vector[1:, :], axis=1)
        tangential = unit_vector[1:, :] - dot[:, None] * unit_vector[:-1, :]
        norm = np.linalg.norm(tangential, axis=1)
        norm[norm == 0.0] = INFINITE_RADIUS
        tangential /= norm[:, None]
        alpha = np.abs(0.5 * np.pi - np.arctan2(dot, vr))
        alpha[alpha == 0.0] = INFINITE_RADIUS**-1.0
        delta_depth = np.diff(full_survey[:, 0])
        radius = delta_depth / alpha

        intervals = {
            "depths": np.r_[full_survey[:, 0]],
            "rad": np.r_[radius, INFINITE_RADIUS],
            "tangential": np.vstack([tangential, np.zeros(3)]),
            "unit_vector": unit_vector,
            "locations": np.r_[
                collar,
                collar
                + np.cumsum(
                    radius[:, None]
                    * (
                        np.sin(alpha)[:, None] * unit_vector[:-1, :]
                        + (1 - np.cos(alpha))[:, None] * tangential
                    ),
                    axis=0,
                ),
            ],
        }

        return intervals

    @staticmethod
    def depths_to_xyz(intervals: dict, depths: np.ndarray) -> np.ndarray:
        """
        Convert survey parameters to x, y, z coordinates.

        :param intervals: Dictionary of intervals parameters.
        :param depths: Array of depth values.
        """
        depths = np.asarray(depths).flatten()
        ind = np.searchsorted(intervals["depths"], depths, side="right") - 1
        dl = depths - intervals["depths"][ind]
        radii = intervals["rad"][ind]
        radii[radii == 0.0] = 1.0

        angle = dl / radii
        locations = intervals["locations"][ind, :] + radii[:, None] * (
            np.sin(angle)[:, None] * intervals["unit_vector"][ind, :]
            + (1 - np.cos(angle))[:, None] * intervals["tangential"][ind, :]
        )
        return locations



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

import logging
from typing import Annotated, Any

from pydantic import BeforeValidator, Field


logger = logging.getLogger(__name__)


def deprecate(value, info):
    """Issue deprecation warning."""
    logger.warning("Skipping deprecated field: %s.", info.field_name)
    return value


Deprecated = Annotated[
    Any,
    Field(exclude=True),
    BeforeValidator(deprecate),
]



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from ..data import Data, DataType
from ..groups import Group, GroupType, PropertyGroup
from ..objects import ObjectBase, ObjectType
from ..shared.entity import Entity
from ..shared.entity_type import EntityType


def str_from_type(entity: Entity | PropertyGroup | EntityType) -> str:
    """
    Convert entity type to geoh5 string standard.

    :param entity: Entity, PropertyGroup or EntityType

    :return: String equivalent.
    """
    if isinstance(entity, Data):
        return "Data"

    if isinstance(entity, Group):
        return "Groups"

    if isinstance(entity, ObjectBase):
        return "Objects"

    if isinstance(entity, PropertyGroup):
        return "PropertyGroups"

    if isinstance(entity, EntityType):
        return "Types"

    raise TypeError(
        "Input value should be of type 'Data', 'Group' or 'ObjectBase'. "
        f"Provided {type(entity)}"
    )


def str_from_subtype(entity_type: EntityType) -> str:
    """
    Convert sub entity type to geoh5 string standard.

    :param entity_type: Sub-class of entity type.

    :return: String equivalent.
    """
    if isinstance(entity_type, DataType):
        return "Data types"
    if isinstance(entity_type, ObjectType):
        return "Object types"
    if isinstance(entity_type, GroupType):
        return "Group types"

    raise TypeError(
        "Input value should be of type 'Data', 'Group' or 'ObjectBase'. "
        f"Provided {type(entity_type)}"
    )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from PIL import Image

from ... import objects
from .base import BaseConversion


if TYPE_CHECKING:
    from ...objects import GeoImage, Grid2D


class GeoImageConversion(BaseConversion):
    """
    Convert a :obj:'geoh5py.objects.geo_image.GeoImage' object.
    """

    @staticmethod
    def convert_to_grid2d_reference(geoimage: GeoImage, grid2d_attributes) -> dict:
        """
        Extract the geographic information from the entity.
        """
        if geoimage.vertices is None or geoimage.default_vertices is None:
            raise AttributeError("GeoImage has no vertices.")

        # get geographic information
        grid2d_attributes["origin"] = np.asarray(
            tuple(geoimage.vertices[3]),
            dtype=[("x", float), ("y", float), ("z", float)],
        )

        grid2d_attributes["u_count"] = geoimage.default_vertices[1, 0].astype(np.int32)
        grid2d_attributes["v_count"] = geoimage.default_vertices[0, 1].astype(np.int32)

        # Compute the distances
        distance_u = np.linalg.norm(geoimage.vertices[2] - geoimage.vertices[3])
        distance_v = np.linalg.norm(geoimage.vertices[0] - geoimage.vertices[3])

        # Now compute the cell sizes
        grid2d_attributes["u_cell_size"] = distance_u / grid2d_attributes["u_count"]
        grid2d_attributes["v_cell_size"] = distance_v / grid2d_attributes["v_count"]

        grid2d_attributes["elevation"] = grid2d_attributes.get("elevation", 0)

        if geoimage.rotation is not None:
            grid2d_attributes["rotation"] = geoimage.rotation

        if geoimage.dip is not None:
            grid2d_attributes["dip"] = geoimage.dip

        return grid2d_attributes

    @staticmethod
    def add_gray_data(values: np.ndarray, output: Grid2D):
        """
        Send the image as gray in the new :obj:'geoh5py.objects.grid2d.Grid2D'.
        :param values: Input image values as an array of int.
        :param output: the new :obj:'geoh5py.objects.grid2d.Grid2D'.
        """
        if values.ndim != 2:
            raise ValueError("To export to gray image, the array must be 2d. ")

        output.add_data(
            data={
                "band[0]": {
                    "values": values[::-1].flatten(),
                    "association": "CELL",
                }
            }
        )

    @staticmethod
    def add_color_data(values: np.ndarray, output: Grid2D):
        """
        Send the image color bands to data.

        :param values: Input image values as an array of int.
        :param output: the new :obj:'geoh5py.objects.grid2d.Grid2D'.
        """
        if values.ndim != 3:
            raise IndexError("To export to color image, the array must be 3d.")

        for ind in range(values.shape[2]):
            output.add_data(
                {
                    f"band[{ind}]": {
                        "values": values[::-1, :, ind].flatten(),
                        "association": "CELL",
                    }
                }
            )

    @staticmethod
    def add_data_2dgrid(geoimage: Image, output: Grid2D):
        """
        Select the type of the image transformation.
        :param geoimage: :obj:'geoh5py.objects.geo_image.GeoImage' object.
        :param output: the new :obj:'geoh5py.objects.grid2d.Grid2D'.
        """
        values = np.asarray(geoimage)

        if values.ndim == 2:
            GeoImageConversion.add_gray_data(values, output)
        else:
            GeoImageConversion.add_color_data(values, output)

    @staticmethod
    def to_grid2d(
        geoimage: GeoImage,
        mode: str | None,
        copy_children=True,
        **grid2d_kwargs,
    ) -> Grid2D:
        """
        Transform the :obj:'geoh5py.objects.image.Image' to a :obj:'geoh5py.objects.grid2d.Grid2D'.
        :param geoimage: :obj:'geoh5py.objects.geo_image.GeoImage' object.
        :param mode: The outgoing image mode option.

        :return: the new :obj:'geoh5py.objects.grid2d.Grid2D'.
        """

        workspace, grid2d_kwargs = GeoImageConversion.validate_workspace(
            geoimage, **grid2d_kwargs
        )
        grid2d_kwargs = GeoImageConversion.verify_kwargs(geoimage, **grid2d_kwargs)
        grid2d_kwargs = GeoImageConversion.convert_to_grid2d_reference(
            geoimage, grid2d_kwargs
        )

        output = objects.Grid2D.create(
            workspace,
            **grid2d_kwargs,
        )

        if geoimage.image is not None:
            image = geoimage.image.copy()

            if mode is not None and mode != image.mode:
                image = image.convert(mode if mode != "GRAY" else "L")

            if copy_children:
                GeoImageConversion.add_data_2dgrid(image, output)

        return output



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..shared.entity_container import EntityContainer
from .group_type import GroupType


if TYPE_CHECKING:
    from ..workspace import Workspace


class Group(EntityContainer):
    """Base Group class"""

    _default_name = "Group"

    def mask_by_extent(self, extent: np.ndarray, inverse: bool = False) -> None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.
        """
        return None

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        **kwargs,
    ):
        """
        Function to copy a group to a different parent entity.

        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param copy_children: (Optional) Create copies of all children entities along with it.
        :param clear_cache: Clear array attributes after copy.
        :param kwargs: Additional keyword arguments to pass to the copy constructor.

        :return entity: Registered Entity to the workspace.
        """
        if parent is None:
            parent = self.parent

        new_entity = parent.workspace.copy_to_parent(
            self, parent, copy_children=False, **kwargs
        )

        if copy_children and new_entity is not None:
            for child in self.children:
                if hasattr(child, "complement") and child.type != "Receivers":
                    continue

                child.copy(
                    parent=new_entity,
                    copy_children=True,
                    clear_cache=clear_cache,
                )

        return new_entity

    def copy_from_extent(
        self,
        extent: np.ndarray,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> Group | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.copy_from_extent`.
        """
        copy_group = self.copy(
            parent=parent,
            clear_cache=clear_cache,
            copy_children=False,
            **kwargs,
        )

        if copy_group is None:
            return None

        if copy_children:
            for child in self.children:
                child.copy_from_extent(
                    extent,
                    parent=copy_group,
                    copy_children=True,
                    clear_cache=clear_cache,
                    inverse=inverse,
                )

            if len(copy_group.children) == 0:
                copy_group.workspace.remove_entity(copy_group, force=True)
                return None

        return copy_group

    @property
    def entity_type(self) -> GroupType:
        return self._entity_type

    @property
    def extent(self) -> np.ndarray | None:
        """
        Geography bounding box of the object.

        :return: shape(2, 3) Bounding box defined by the bottom South-West and
            top North-East coordinates.
        """
        extents = []
        for child in self.children:
            if child.extent is not None:
                extents.append(child.extent)

        if len(extents) > 0:
            extents = np.vstack(extents)
            return np.vstack(
                [
                    np.min(extents, axis=0),
                    np.max(extents, axis=0),
                ]
            )

        return None

    @classmethod
    def find_or_create_type(cls, workspace: Workspace, **kwargs) -> GroupType:
        kwargs["entity_class"] = cls
        return GroupType.find_or_create(workspace, **kwargs)

    def validate_entity_type(self, entity_type: GroupType | None) -> GroupType:
        """
        Validate the entity type.
        """
        if not isinstance(entity_type, GroupType):
            raise TypeError(
                f"Input 'entity_type' must be of type {GroupType}, not {type(entity_type)}"
            )

        if entity_type.name == "Entity":
            entity_type.name = self.name

        return entity_type



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from .object_base import ObjectBase


if TYPE_CHECKING:
    from numpy import ndarray


class NoTypeObject(ObjectBase):
    """
    Generic Data object without a registered type
    """

    _TYPE_UID = uuid.UUID(
        fields=(0x849D2F3E, 0xA46E, 0x11E3, 0xB4, 0x01, 0x2776BDF4F982)
    )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from .grid_object import GridObject
from .object_base import ObjectBase


class DrapeModel(ObjectBase):
    """
    Drape (curtain) model object made up of layers and prisms.

    :param layers: Array of layers in the drape model organized into blocks
        representing each prism in the model.
    :param prisms: Array detailing the assembly of
        :obj:`geoh5py.objects.drape_model.DrapeModel.layers` within the trace
        of the drape model.
    """

    _TYPE_UID = uuid.UUID("{C94968EA-CF7D-11EB-B8BC-0242AC130003}")
    __LAYERS_DTYPE = np.dtype([("I", "<i4"), ("K", "<i4"), ("Bottom elevation", "<f8")])
    __PRISM_DTYPE = np.dtype(
        [
            ("Top easting", "<f8"),
            ("Top northing", "<f8"),
            ("Top elevation", "<f8"),
            ("First layer", "<i4"),
            ("Layer count", "<i4"),
        ]
    )

    def __init__(
        self,
        layers: np.ndarray | list | tuple | None = None,
        prisms: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        self._centroids: np.ndarray | None = None
        self._layers: np.ndarray = self.validate_layers(layers)
        self._prisms: np.ndarray = self.validate_prisms(prisms)

        super().__init__(**kwargs)

    @property
    def centroids(self) -> np.ndarray:
        """
        Cell center locations in world coordinates, shape(*, 3).

        .. code-block:: python

            centroids = [
                [x_1, y_1, z_1],
                ...,
                [x_N, y_N, z_N]
            ]
        """
        if getattr(self, "_centroids", None) is None:
            self._centroids = np.vstack(
                [
                    np.ones((int(val), 3)) * self.prisms[ii, :3]
                    for ii, val in enumerate(self.prisms[:, 4])
                ]
            )
            tops = np.hstack(
                [
                    np.r_[
                        cells[2],
                        self.layers[int(cells[3]) : int(cells[3] + cells[4] - 1), 2],
                    ]
                    for cells in self.prisms.tolist()
                ]
            )
            self._centroids[:, 2] = (tops + self.layers[:, 2]) / 2.0

        return self._centroids

    @property
    def layers(self) -> np.ndarray:
        """
        Layers in the drape model organized into blocks representing each prism in the model:

        X (prism index), K (depth index), elevation (cell bottom))

        .. code-block:: python

            layers = [
                [x_1, k_1, z_11],
                [x_1, k_2, z_12],
                ...
                [x_1, k_N, z_1N],
                .
                .
                .
                [x_M, k_1, z_M1],
                [x_M, k_2, z_M2],
                ...
                [x_M, k_N, z_MM]
            ]
        """
        if self._layers is None and self.on_file:
            self._layers = self.workspace.fetch_array_attribute(self, "layers")

        return np.asarray(self._layers.tolist())

    @property
    def n_cells(self) -> int:
        return self.layers.shape[0]

    @property
    def prisms(self) -> np.ndarray:
        """
        Array detailing the assembly of :obj:
        `geoh5py.objects.drape_model.Drapemodel.layers` within the trace
        of the drape model.

        Columns: Easting, Northing, Elevation (top),
        layer index (first), layer count.

        .. code-block:: python

            prisms = [
                [e_1, n_1, z_1, l_1, c_1],
                ...,
                [e_N, n_N, z_N, l_N, c_N]
            ]

        """
        if self._prisms is None and self.on_file:
            self._prisms = self.workspace.fetch_array_attribute(self, "prisms")

        return np.array(self._prisms.tolist())

    @classmethod
    def validate_prisms(cls, values: np.ndarray | list | tuple | None) -> np.ndarray:
        """
        Validate and format type of prisms array.

        :param values: Array of prisms as defined by
            :obj:`~geoh5py.objects.drape_model.DrapeModel.prisms`.
        """
        if isinstance(values, (list, tuple)):
            values = np.array(values, ndmin=2)

        if not isinstance(values, np.ndarray):
            raise TypeError(
                "Attribute 'prisms' must be a list, tuple or numpy array. "
                f"Object of type {type(values)} provided."
            )

        if np.issubdtype(values.dtype, np.number):
            if values.shape[1] != 5:
                raise ValueError(
                    "Array of 'prisms' must be of shape (*, 5). "
                    f"Array of shape {values.shape} provided."
                )

            values = np.asarray(
                np.core.records.fromarrays(
                    values.T.tolist(),
                    dtype=cls.__PRISM_DTYPE,
                )
            )

        if values.dtype != cls.__PRISM_DTYPE:
            raise ValueError(
                f"Array of 'prisms' must be of dtype = {cls.__PRISM_DTYPE}"
            )

        return values

    @classmethod
    def validate_layers(cls, values: np.ndarray | list | tuple | None) -> np.ndarray:
        """
        Validate and format type of layers array.

        :param values: Array of layers as defined by
            :obj:`~geoh5py.objects.drape_model.DrapeModel.layers`.
        """
        if isinstance(values, (list, tuple)):
            values = np.array(values, ndmin=2)

        if not isinstance(values, np.ndarray):
            raise TypeError(
                "Attribute 'layers' must be a list, tuple or numpy array. "
                f"Object of type {type(values)} provided."
            )

        if np.issubdtype(values.dtype, np.number):
            if values.shape[1] != 3:
                raise ValueError(
                    "Array of 'layers' must be of shape (*, 3). "
                    f"Array of shape {values.shape} provided."
                )

            if any(np.diff(np.unique(values[:, 0])) != 1):
                msg = "Prism index (first column) must be monotonically increasing."
                raise ValueError(msg)

            values = np.asarray(
                np.core.records.fromarrays(
                    values.T.tolist(),
                    dtype=cls.__LAYERS_DTYPE,
                )
            )

        if values.dtype != cls.__LAYERS_DTYPE:
            raise ValueError(
                f"Array of 'layers' must be of dtype = {cls.__LAYERS_DTYPE}"
            )

        return values


DrapeModel.validate_mask = GridObject.validate_mask  # type: ignore
DrapeModel.validate_cell_mask = GridObject.validate_cell_mask  # type: ignore



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

from .base import Group


class AirborneTheme(Group):
    """The type for a INTEGRATOR Airborne Theme."""

    _TYPE_UID = uuid.UUID("{3d0e8578-7764-48cf-8db8-6c83d6411762}")
    _default_name = "Airborne Theme"


class BlastingDataSet(Group):
    """The type for a INTEGRATOR Blasting Data Set."""

    _TYPE_UID = uuid.UUID("{f55d8ae4-3829-11e4-a70e-fcddabfddab1}")
    _default_name = "Blasting Data Set"


class BlastingTheme(Group):
    """The type for a INTEGRATOR Blasting Theme."""

    _TYPE_UID = uuid.UUID("{e2040afa-3829-11e4-a70e-fcddabfddab1}")
    _default_name = "Blasting Theme"


class EarthModelsTheme(Group):
    """The type for a INTEGRATOR Earth Models Theme."""

    _TYPE_UID = uuid.UUID("{adee3b2a-3829-11e4-a70e-fcddabfddab1}")
    _default_name = "Earth Models Theme"


class GeochemistryMineralogyTheme(Group):
    """The type for a INTEGRATOR Geochemistry & Mineralogy Theme."""

    _TYPE_UID = uuid.UUID("{ed00094f-3da1-485f-8c4e-b52f6f171ea4}")
    _default_name = "Geochemistry & Mineralogy Theme"


class GeochemistryMineralogyDataSet(Group):
    """The type for a INTEGRATOR Geochemistry & Mineralogy DataSet."""

    _TYPE_UID = uuid.UUID("{72f29283-a4f6-4fc0-a1a8-1417ce5fcbec}")
    _default_name = "Geochemistry & Mineralogy DataSet"


class GeophysicsTheme(Group):
    """The type for a INTEGRATOR Geophysics Theme."""

    _TYPE_UID = uuid.UUID("{151778d9-6cc0-4e72-ba08-2a80a4fb967f}")
    _default_name = "Geophysics Theme"


class GroundTheme(Group):
    """The type for a INTEGRATOR Ground Theme."""

    _TYPE_UID = uuid.UUID("{47d6f059-b56a-46c7-8fc7-a0ded87360c3}")
    _default_name = "Ground Theme"


class IncidentsTheme(Group):
    """The type for a INTEGRATOR Incident Theme."""

    _TYPE_UID = uuid.UUID("{136cb431-c7d2-4992-a5ab-46a6e16b6726}")
    _default_name = "Incidents Theme"


class IncidentDataSet(Group):
    """The type for a INTEGRATOR Incident Data Set."""

    _TYPE_UID = uuid.UUID("{016dfd26-7d9b-49a6-97d8-cb31c37e404b}")
    _default_name = "Incident Data Set"


class IntegratorGroup(Group):
    """The type for a INTEGRATOR group."""

    _TYPE_UID = uuid.UUID("{61449477-3833-11e4-a7fb-fcddabfddab1}")
    _default_name = "Geoscience INTEGRATOR"


class IntegratorMaps(Group):
    """The type for a INTEGRATOR Maps Group."""

    _TYPE_UID = uuid.UUID("{4d65f8c3-a015-4c01-b411-412c0f4f0884}")
    _default_name = "Maps Group"


class IntegratorMapFile(Group):
    """The type for a INTEGRATOR Map File Group."""

    _TYPE_UID = uuid.UUID("{1f684938-2baf-4a01-ac71-e50c30cc0685}")
    _default_name = "Geoscience INTEGRATOR Map File Group"


class IntegratorNeighbourhoods(Group):
    """The type for a INTEGRATOR Neighbourhoods Group."""

    _TYPE_UID = uuid.UUID("{2a5b7faa-41d1-4437-afac-934933eae6eb}")
    _default_name = "Neighbourhoods"


class IntegratorProject(Group):
    """The type for a INTEGRATOR group."""

    _TYPE_UID = uuid.UUID("{56f6f03e-3833-11e4-a7fb-fcddabfddab1}")
    _default_name = "Geoscience INTEGRATOR Project"


class MineProductionTheme(Group):
    """The type for a INTEGRATOR Mine Production Theme."""

    _TYPE_UID = uuid.UUID("{7508bc11-3829-11e4-9cce-fcddabfddab1}")
    _default_name = "Mine Production Theme"


class ObservationPointsTheme(Group):
    """The type for a INTEGRATOR Observation Points Theme."""

    _TYPE_UID = uuid.UUID("{f65e521c-a763-427b-97bf-d0b4e5689e0d}")
    _default_name = "Observation Points Theme"


class QueryGroup(Group):
    """The type for a INTEGRATOR Query Group."""

    _TYPE_UID = uuid.UUID("{85756113-592a-4088-b374-f32c8fac37a2}")
    _default_name = "Query Group"


class RockPropertiesTheme(Group):
    """The type for a INTEGRATOR Rock Properties Theme."""

    _TYPE_UID = uuid.UUID("{cbeb3920-a1a9-46f8-ab2b-7dfdf79c8a00}")
    _default_name = "Rock Properties Theme"


class SamplesTheme(Group):
    """The type for a INTEGRATOR Samples Theme."""

    _TYPE_UID = uuid.UUID("{1cde9996-cda7-40f0-8c20-faeb4e926748}")
    _default_name = "Samples Theme"



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import cast

import numpy as np

from ..shared.utils import find_unique_name
from .data import Data
from .data_type import DataType, GeometricDataValueMapType, ReferenceDataType
from .geometric_data import GeometricDataConstants
from .integer_data import IntegerData
from .primitive_type_enum import PrimitiveTypeEnum
from .reference_value_map import ReferenceValueMap


class ReferencedData(IntegerData):
    """
    Reference data described by indices and associated strings.
    """

    def __init__(self, **kwargs):
        self._data_maps = None

        super().__init__(**kwargs)

    def copy(
        self,
        parent=None,
        *,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ) -> Data:
        """
        Overwrite the copy method to ensure that the uid is set to None and
        transfer the GeometricDataValueMapType to the new entity type.
        """
        kwargs["uid"] = None

        new_data = cast(
            ReferencedData,
            super().copy(
                parent=parent,
                clear_cache=clear_cache,
                mask=mask,
                omit_list=["_metadata", "_data_maps"],
                **kwargs,
            ),
        )

        # Always overwrite the entity type name to protect the GeometricDataValueMapType
        new_data.entity_type.name = find_unique_name(
            self.entity_type.name,
            [tp.name for tp in self.workspace.types if isinstance(tp, DataType)],
        )

        if self.data_maps is not None:
            for name, child in self.data_maps.items():
                new_data.add_data_map(name, child.entity_type.value_map.map)

        return new_data

    @property
    def data_maps(self) -> dict[str, GeometricDataConstants] | None:
        """
        A reference dictionary mapping properties to numpy arrays.
        """
        if self._data_maps is None and self.metadata is not None:
            data_maps = {}
            for name, uid in self.metadata.items():
                child = self.workspace.get_entity(uid)[0]
                if isinstance(child, GeometricDataConstants) and isinstance(
                    child.entity_type, GeometricDataValueMapType
                ):
                    data_maps[name] = child

            if data_maps:
                self._data_maps = data_maps

        return self._data_maps

    @data_maps.setter
    def data_maps(self, data_map: dict[str, GeometricDataConstants] | None):
        if data_map is not None:
            if not isinstance(data_map, dict):
                raise TypeError("Property maps must be a dictionary")
            for key, val in data_map.items():
                if not isinstance(val, GeometricDataConstants):
                    raise TypeError(
                        f"Property maps value for '{key}' must be a 'GeometricDataConstants'."
                    )

                if (
                    not isinstance(val.entity_type, GeometricDataValueMapType)
                    or val.entity_type.value_map is None
                ):
                    raise ValueError(
                        f"Property maps value for '{key}' must have a "
                        "'GeometricDataValueMapType' entity type and a 'value_map' assigned."
                    )

            self.metadata = {child.name: child.uid for child in data_map.values()}

        self._data_maps = data_map

        if self.on_file:
            self.workspace.update_attribute(self, "data_map")

    @property
    def entity_type(self) -> ReferenceDataType:
        """
        The associated reference data type.
        """
        return self._entity_type

    @entity_type.setter
    def entity_type(self, data_type: ReferenceDataType):
        if not isinstance(data_type, ReferenceDataType):
            raise TypeError("entity_type must be of type ReferenceDataType")

        self._entity_type = data_type

        if self.on_file:
            self.workspace.update_attribute(self, "entity_type")

    @property
    def mapped_values(self) -> np.ndarray:
        """
        The values mapped from the reference data.
        """
        if self.value_map is None:
            raise ValueError("Entity type must have a value map.")

        return self.value_map.map_values(self.values)

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.REFERENCED

    @property
    def value_map(self) -> ReferenceValueMap | None:
        """
        Pointer to the :obj:`data.data_type.DataType.value_map`
        """
        return self.entity_type.value_map

    def remove_data_map(self, name: str):
        """
        Remove a data map from the list of children.

        :param name: The name of the data map to remove.
        """

        if self.data_maps is None or name not in self.data_maps:
            return

        child = self.data_maps[name]
        child.allow_delete = True
        self.workspace.remove_entity(child)
        self.workspace.remove_entity(child.entity_type)

        del self.data_maps[name]
        self.data_maps = self._data_maps

    def add_data_map(self, name: str, data: np.ndarray | dict):
        """
        Add a data map to the value map.

        :param name: The name of the data map.
        :param data: The data map to add.
        """
        value_map = self.data_maps or {}

        if name in value_map:
            raise KeyError(f"Data map '{name}' already exists.")

        if not isinstance(data, dict | np.ndarray):
            raise TypeError("Data map values must be a numpy array or dict")

        if self.entity_type.value_map is None:
            raise ValueError("Entity type must have a value map.")

        reference_data = ReferenceValueMap(data, name=name)
        # TODO: Enforce that the keys of the data map are a subset
        #  of the value map keys once GA changes its behavior
        # if not set(reference_data.map["Key"]).issubset(
        #     set(self.entity_type.value_map.map["Key"])
        # ):
        #     raise KeyError("Data map keys must be a subset of the value map keys.")
        #

        data_type = GeometricDataValueMapType(
            self.workspace,
            value_map=reference_data,
            parent=self.parent,
            name=self.entity_type.name + f": {name}",
        )
        geom_data = self.parent.add_data(
            {
                name: {
                    "association": self.association,
                    "entity_type": data_type,
                }
            }
        )
        value_map[name] = geom_data
        self.data_maps = value_map



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=R0904

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any
from uuid import UUID
from warnings import warn

import numpy as np

from ..data import (
    CommentsData,
    Data,
    DataAssociationEnum,
    GeometricDataConstants,
    VisualParameters,
)
from ..groups.property_group import GroupTypeEnum, PropertyGroup
from ..shared import Entity
from ..shared.conversion import BaseConversion
from ..shared.entity_container import EntityContainer
from ..shared.utils import (
    array_is_colour,
    box_intersect,
    clear_array_attributes,
    mask_by_extent,
    str2uuid,
)
from .object_type import ObjectType


if TYPE_CHECKING:  # pragma: no cover
    from ..workspace import Workspace

logger = logging.getLogger(__name__)


class ObjectBase(EntityContainer):
    """
    Object base class.

    :param last_focus: Object visible in camera on start.
    """

    _attribute_map: dict = EntityContainer._attribute_map.copy()
    _attribute_map.update(
        {"Last focus": "last_focus", "PropertyGroups": "property_groups"}
    )
    _converter: type[BaseConversion] | None = None

    def __init__(self, last_focus: str = "None", **kwargs):
        self._property_groups: list[PropertyGroup] | None = None
        self._visual_parameters: VisualParameters | None = None
        self._comments: CommentsData | None = None

        self.last_focus = last_focus

        super().__init__(**kwargs)

    def _remove_children_values(
        self,
        indices: list[int] | np.ndarray,
        association: DataAssociationEnum,
        clear_cache: bool = False,
    ):
        """
        Remove values from children data objects.

        :param indices: The indices to remove.
        :param association: The association of the data to remove.
        :param clear_cache: Clear the cache of the children.
        """
        for child in self.children:
            if (
                isinstance(child, Data)
                and isinstance(child.values, np.ndarray)
                and child.association == association
            ):
                child.values = np.delete(child.values, indices, axis=0)

                if child.values.size == 0:
                    child.values = None

                if child.on_file:
                    child.workspace.update_attribute(child, "values")

                if clear_cache:
                    clear_array_attributes(child)

    def add_children(
        self, children: Entity | PropertyGroup | list[Entity | PropertyGroup]
    ):
        """
        :param children: a list of entity to add as children.
        """
        property_groups = self._property_groups or []

        if not isinstance(children, list):
            children = [children]

        children_uids = {child.uid: child for child in self._children}

        for child in children:
            if (
                isinstance(child, (Data, PropertyGroup))
                and child.uid not in children_uids
            ):
                self._children.append(child)
                if isinstance(child, PropertyGroup):
                    property_groups.append(child)
                elif hasattr(child, "parent") and child.parent != self:
                    child.parent = self
            else:
                warn(f"Child {child} is not valid or already exists.")

        if property_groups:
            self._property_groups = property_groups

    def add_data(
        self,
        data: dict,
        property_group: str | PropertyGroup | None = None,
        compression: int = 5,
        **kwargs,
    ) -> Data | list[Data]:
        """
        Create a Data from dictionary of name and arguments.
        The provided arguments can be any property of the target Data class.

        :param data: Dictionary of data to be added to the object, e.g.
        :param property_group: Name or :obj:`~geoh5py.groups.property_group.PropertyGroup`.
        :param compression: Compression level for data.

        .. code-block:: python

            data = {
                "data_A": {
                    'values': [v_1, v_2, ...],
                    'association': 'VERTEX'
                    },
                "data_B": {
                    'values': [v_1, v_2, ...],
                    'association': 'CELLS'
                    },
            }

        :return: List of new Data objects.
        """
        if not isinstance(data, dict):
            raise TypeError(
                f"Input 'data' must be of type {dict}. Got {type(data)} instead."
            )

        property_groups: dict[PropertyGroup | None, list[Data]] = {}
        data_objects = []
        for name, attr in data.items():
            if not isinstance(attr, dict):
                raise TypeError(
                    f"Given value to data {name} should of type {dict}. "
                    f"Type {type(attr)} given instead."
                )

            attr, validate_property_group = self.validate_association(
                {**attr, "name": name}, property_group=property_group, **kwargs
            )

            data_object = self.workspace.create_entity(
                Data,
                entity={
                    "parent": self,
                    **{
                        key: val
                        for key, val in attr.items()
                        if key not in ["parent", "entity_type", "type"]
                    },
                },
                entity_type=self.workspace.validate_data_type(attr, attr.get("values")),
                compression=compression,
            )

            # change the visual parameters if the data object is a visual parameter
            if isinstance(data_object, VisualParameters):
                self.visual_parameters = data_object

            property_groups.setdefault(validate_property_group, []).append(data_object)
            data_objects.append(data_object)

        for proper_group, data_associated in property_groups.items():
            if proper_group is not None:
                self.add_data_to_group(
                    data_associated,  # type: ignore
                    proper_group,
                    property_group_type=GroupTypeEnum.find_type(data_associated),
                )

        # TODO: Legacy re-sorting for old drillhole format
        self.post_processing()

        if len(data_objects) == 1:
            return data_objects[0]

        return data_objects

    def add_data_to_group(
        self,
        data: list[Data | UUID | str] | Data | UUID | str,
        property_group: str | PropertyGroup,
        **kwargs,
    ) -> PropertyGroup:
        """
        Append data children to a :obj:`~geoh5py.groups.property_group.PropertyGroup`
        All given data must be children of the parent object.

        :param data: The name, the uid or the object to add itself, pass as a list or single object.
        :param property_group: The name or the object of the property group;
            a new one will be created if not found.
        :param kwargs: Additional keyword arguments to create a property group.

        :return: The target property group.
        """
        if isinstance(property_group, str):
            property_group = self.fetch_property_group(
                name=property_group, properties=data, **kwargs
            )

        if isinstance(property_group, PropertyGroup):
            property_group.add_properties(data)
            return property_group

        raise TypeError(
            "Property group must be of type PropertyGroup or str; "
            f"got {type(property_group)} instead."
        )

    def add_default_visual_parameters(self):
        """
        Create a default visual parameters to the object.
        """
        if self.visual_parameters is not None:
            raise UserWarning("Visual parameters already exist.")

        self._visual_parameters = self.workspace.create_entity(
            Data,  # type: ignore
            save_on_creation=True,
            entity={
                "name": "Visual Parameters",
                "parent": self,
                "association": "OBJECT",
            },
            entity_type={"name": "XmlData", "primitive_type": "TEXT"},
        )

        return self._visual_parameters

    @property
    def converter(self) -> Any:
        """
        :return: The converter for the object.
        """
        return self._converter

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Function to copy an entity to a different parent entity.

        :param parent: New parent for the copied object.
        :param copy_children: Copy children entities.
        :param clear_cache: Clear cache of data values.
        :param mask: Array of indices to sub-sample the input entity.
        :param kwargs: Additional keyword arguments.

        :return: New copy of the input entity.
        """
        new_object = self.workspace.copy_to_parent(
            self,
            parent or self.parent,
            clear_cache=clear_cache,
            **kwargs,
        )

        mask = self.validate_mask(mask)

        if copy_children:
            children_map = {}
            for child in self.children:
                if isinstance(child, PropertyGroup | GeometricDataConstants):
                    continue

                child_copy = child.copy(
                    parent=new_object,
                    clear_cache=clear_cache,
                    mask=mask,
                )

                children_map[child.uid] = child_copy.uid

            if self.property_groups:
                self.workspace.copy_property_groups(
                    new_object, self.property_groups, children_map
                )
                new_object.workspace.update_attribute(new_object, "property_groups")

        return new_object

    def create_property_group(
        self,
        name=None,
        property_group_type: GroupTypeEnum | str = GroupTypeEnum.MULTI,
        **kwargs,
    ) -> PropertyGroup:
        """
        Create a new :obj:`~geoh5py.groups.property_group.PropertyGroup`.

        :param name: Name of the new property group.
        :param property_group_type: Type of property group.
        :param kwargs: Any arguments taken by the
            :obj:`~geoh5py.groups.property_group.PropertyGroup` class.

        :return: A new :obj:`~geoh5py.groups.property_group.PropertyGroup`
        """
        prop_group = PropertyGroup(
            self, name=name, property_group_type=property_group_type, **kwargs
        )

        return prop_group

    @property
    def entity_type(self) -> ObjectType:
        """
        :obj:`~geoh5py.shared.entity_type.EntityType`: Object type.
        """
        return self._entity_type

    @property
    def extent(self) -> np.ndarray | None:
        """
        Geography bounding box of the object.

        :return: Bounding box defined by the bottom South-West and
            top North-East coordinates,  shape(2, 3).
        """
        if self.locations is None:
            return None

        return np.c_[self.locations.min(axis=0), self.locations.max(axis=0)].T

    def fetch_property_group(self, name=None, uid=None, **kwargs) -> PropertyGroup:
        """
        Find or create a PropertyGroup from given name and properties.

        :param name: Name of the property group.
        :param uid: Unique identifier for the property group.
        :param kwargs: Any arguments taken by the
            :obj:`~geoh5py.groups.property_group.PropertyGroup` class.

        :return: A new or existing :obj:`~geoh5py.groups.property_group.PropertyGroup`
        """
        prop_group = None
        if name is not None or uid is not None:
            prop_group = self.get_property_group(uid or name)[0]

        if prop_group is None:
            prop_group = self.create_property_group(name=name, uid=uid, **kwargs)

        return prop_group

    def find_association(self, values: np.ndarray) -> str:
        """
        Find the association based on a value shape.

        :param values: The values to check.

        :return: The name of the association.
        """
        if isinstance(values, np.ndarray):
            values = values if array_is_colour(values) else values.ravel()

            if values.shape[0] == getattr(self, "n_cells", None):
                return "CELL"
            if values.shape[0] == getattr(self, "n_vertices", None):
                return "VERTEX"

        return "OBJECT"

    def find_or_create_property_group(
        self, name=None, uid=None, **kwargs
    ) -> PropertyGroup:
        """
        Find or create a PropertyGroup from given name and properties.
        """
        warn(
            "The 'find_and_create_property_group' will be deprecated. "
            "Use fetch_property_group instead.",
            DeprecationWarning,
        )
        return self.fetch_property_group(name=name, uid=uid, **kwargs)

    @classmethod
    def find_or_create_type(cls, workspace: Workspace, **kwargs) -> ObjectType:
        """
        Find or create a type instance for a given object class.

        :param workspace: Target workspace.
        :param kwargs: Keyword arguments for the ObjectType.

        :return: The ObjectType instance for the given object class.
        """
        kwargs["entity_class"] = cls
        return ObjectType.find_or_create(workspace, **kwargs)

    def get_property_group(self, name: UUID | str) -> list[PropertyGroup] | list[None]:
        """
        Get a child PropertyGroup by name.

        :param name: the reference of the property group to get.

        :return: A list of children Data objects
        """
        if self._property_groups is None:
            return [None]

        entities = []
        for child in self._property_groups:
            if (isinstance(name, UUID) and child.uid == name) or child.name == name:
                entities.append(child)

        if len(entities) == 0:
            return [None]

        return entities

    def get_data(self, name: str | UUID) -> list[Data]:
        """
        Get the children associated with a given name.

        :param name: Name or UUID of the target child data

        :return: A list of children Data objects
        """
        return [child for child in self.get_entity(name) if isinstance(child, Data)]

    def get_data_list(self, attribute: str = "name") -> list[Any]:
        """
        Get a list of a specific attribute of the data children.

        :param attribute: The attribute to return from the data objects.

        :return: List of names of data associated with the object.
        """
        return sorted(
            [
                getattr(child, attribute)
                for child in self.children
                if isinstance(child, Data)
            ]
        )

    @property
    def last_focus(self) -> str:
        """
        The name of the object visible in the camera on start.
        """
        return self._last_focus

    @last_focus.setter
    def last_focus(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Attribute 'last_focus' must be a string")

        self._last_focus = value

    def load_children_values(self):
        """
        Load the values of the children in memory.
        """
        for child in self.children:
            _ = getattr(child, "values", None)

    @property
    def locations(self) -> np.ndarray | None:
        """
        Exposes the vertices or centroids of the object.
        """
        out = None
        if hasattr(self, "vertices"):
            out = self.vertices
        if hasattr(self, "centroids"):
            out = self.centroids

        return out

    def mask_by_extent(
        self, extent: np.ndarray, inverse: bool = False
    ) -> np.ndarray | None:
        if self.extent is None or not box_intersect(self.extent, extent):
            return None

        return mask_by_extent(self.locations, extent, inverse=inverse)

    def post_processing(self):
        """
        Post-processing function to be called after adding data.
        """

    @property
    def property_groups(self) -> list[PropertyGroup] | None:
        """
        List of the property groups associated with the object.
        """
        return self._property_groups

    def reference_to_data(self, data: str | Data | UUID) -> Data:
        """
        Convert a reference to a Data object.

        :param data: The data to convert.
            It can be the name, the uuid or the data itself.

        :return: The data object.
        """
        data = str2uuid(data)

        if isinstance(data, Data):
            if self != data.parent:
                raise ValueError(
                    f"Data '{data.name}' parent ({data.parent}) "
                    f"does not match group parent ({self})."
                )

        if isinstance(data, (str, UUID)):
            data_: list = self.get_data(data)
            if len(data_) == 0 and isinstance(data, UUID):
                data_temp = self.workspace.load_entity(data, "data", self)
                data_ = [] if data_temp is None else [data_temp]
            if len(data_) == 0:
                raise ValueError(f"Data '{data}' not found in parent {self}")
            if len(data_) > 1:
                raise ValueError(f"Multiple data '{data}' found in parent {self}")
            data = data_[0]

        if not isinstance(data, Data):
            raise TypeError(
                f"Data must be of type Data, UUID or str. Provided {type(data)}"
            )

        return data

    def remove_children(self, children: list[Entity | PropertyGroup]):
        """
        Remove children from the list of children entities.

        :param children: List of entities

        .. warning::
            Removing a child entity without re-assigning it to a different
            parent may cause it to become inactive. Inactive entities are removed
            from the workspace by
            :func:`~geoh5py.shared.weakref_utils.remove_none_referents`.
        """
        if not isinstance(children, list):
            children = [children]

        for child in children.copy():
            if child not in self._children:
                warn(f"Child {child} not found in parent {self}.")
                children.remove(child)
                continue
            if (
                isinstance(child, PropertyGroup)
                and self._property_groups
                and child in self._property_groups
            ):
                self._property_groups.remove(child)
            elif isinstance(child, Data):
                self.remove_data_from_groups(child)

            self._children.remove(child)

        self.workspace.remove_children(self, children)

    def validate_association(self, attributes, property_group=None, **_):
        """
        Get a dictionary of attributes and validate the data 'association' keyword.

        :param attributes: Dictionary of attributes provided for the data.
        :param property_group: Property group to associate the data with.
        """
        if attributes.get("association") is not None or "values" not in attributes:
            return attributes, property_group

        attributes["association"] = self.find_association(attributes["values"])

        return attributes, property_group

    def validate_mask(self, mask: np.ndarray | None) -> np.ndarray | None:
        """
        Validate the mask array.
        """
        if mask is not None:
            logger.warning("Masking is not supported for %s objects.", type(self))

        return mask

    def remove_data_from_groups(
        self, data: list[Data | UUID | str] | Data | UUID | str
    ):
        """
        Remove data children to all
        :obj:`~geoh5py.groups.property_group.PropertyGroup` of the object.

        :param data: The name, the uid or the object to remove itself,
            pass as a list or single object.
        """
        if not isinstance(data, list):
            data = [data]

        if not self._property_groups:
            return

        for property_group in self._property_groups:
            property_group.remove_properties(data)

    def validate_entity_type(self, entity_type: ObjectType) -> ObjectType:
        """
        Validate the entity type.
        """
        if not isinstance(entity_type, ObjectType):
            raise TypeError(
                f"Input 'entity_type' must be of type {ObjectType}, not {type(entity_type)}"
            )

        if entity_type.name == "Entity":
            entity_type.name = self.name
            entity_type.description = self.name

        return entity_type

    @property
    def visual_parameters(self) -> VisualParameters | None:
        """
        The visual parameters of the object.
        """
        if self._visual_parameters is None:
            for child in self.children:
                if isinstance(child, VisualParameters):
                    self._visual_parameters = child
                    break

        return self._visual_parameters

    @visual_parameters.setter
    def visual_parameters(self, value: VisualParameters | None):
        if not isinstance(value, VisualParameters | None):
            raise TypeError(
                f"Input 'visual_parameters' must be of type {VisualParameters}, "
                f"not {type(value)}"
            )
        if self.visual_parameters is not None:
            self.remove_children([self.visual_parameters])

        self._visual_parameters = value
